"""
Base classes for agents and nodes.
"""

from __future__ import annotations

import json
import logging
import time
from abc import abstractmethod
from typing import Any, Generator, Generic

import aiohttp
from pydantic import BaseModel, ConfigDict, Field, SerializeAsAny
from termcolor import colored
from typing_extensions import Self

from tapeagents.core import (
    Action,
    AgentEvent,
    AgentStep,
    AnnotatorTapeType,
    LLMOutputParsingFailureAction,
    MakeObservation,
    ObservationMakerTapeType,
    PartialStep,
    Prompt,
    Step,
    Tape,
    TapeMetadata,
    TapeType,
    Thought,
    TrainingText,
)
from tapeagents.llms import LLM, LLMCall, LLMEvent, LLMOutput, LLMStream, TrainableLLM
from tapeagents.observe import observe_llm_call
from tapeagents.tool_calling import ToolSpec
from tapeagents.view import TapeViewStack

DEFAULT = "default"

logger = logging.getLogger(__name__)


class AgentStream(Generic[TapeType]):
    """
    A wrapper around a generator that produces AgentEvents, representing the result of an agent run.

    The generator can be iterated over to get the events, or the final tape can be extracted with get_final_tape.
    Support iterable protocol and generator protocol.

    Attributes:
        generator (Generator[AgentEvent[TapeType], None, None]): The generator that produces AgentEvents.
    """

    def __init__(self, generator: Generator[AgentEvent[TapeType], None, None]):
        self.generator = generator

    def __iter__(self):
        return self.generator

    def __next__(self) -> AgentEvent[TapeType]:
        return next(self.generator)

    def get_final_tape(self) -> TapeType:
        """
        Retrieve the final tape from the agent's events.

        Iterates through the events of the agent and returns the final tape
        if it is found. If no final tape is produced by the agent, a ValueError
        is raised.

        Returns:
            TapeType: The final tape produced by the agent.

        Raises:
            ValueError: If the agent did not produce a final tape.
        """
        for event in self:
            if event.final_tape:
                return event.final_tape
        raise ValueError("Agent didn't produce final tape")

    def get_steps(self) -> Generator[Step, None, None]:
        """
        Generator function that yields steps from events.

        Yields:
            Step: The step associated with each event that has a step.
        """
        for event in self:
            if event.step:
                yield event.step


class Node(BaseModel):
    """
    A node in the agent, atomic unit of the agent's behavior.

    The agent chooses which node to run based on the current tape.
    The node has a name and contains 2 main functions:

    - make a prompt out of the tape
    - generate steps out of the received llm output

    Attributes:
        name (str): The name of the node. Defaults to an empty string.
    """

    name: str = ""
    llm: str = DEFAULT

    def model_post_init(self, __context: Any) -> None:
        if not self.name:
            self.name = self.__class__.__name__.split("[")[0]  # class name without type variables

    def make_prompt(self, agent: Any, tape: Tape) -> Prompt:
        """
        Creates a prompt for the given agent and tape.

        Args:
            agent (Any): The agent for which the prompt is being created.
            tape (Tape): The tape associated with the agent.

        Returns:
            Prompt: The generated prompt.
        """
        return Prompt()

    def generate_steps(
        self, agent: Any, tape: Tape, llm_stream: LLMStream
    ) -> Generator[Step | PartialStep, None, None]:
        """
        Generates steps for the given agent, tape, and LLM stream.

        Args:
            agent (Any): The agent for which steps are to be generated.
            tape (Tape): The tape object containing relevant data.
            llm_stream (LLMStream): The LLM stream to be used for generating steps.

        Yields:
            Union[Step, PartialStep]: The generated steps or partial steps.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError("Node must implement generate_steps")

    def make_llm_output(self, agent: Any, tape: Tape, index: int) -> LLMOutput:
        """
        Generates an LLMOutput object for a given agent and tape at a specified index.

        Args:
            agent (Any): The agent for which the LLMOutput is being generated.
            tape (Tape): The tape containing the steps.
            index (int): The index of the step in the tape from which to generate the output.

        Returns:
            LLMOutput: An object containing the role and content for the LLM output.
        """
        return LLMOutput(role="assistant", content=tape.steps[index].content)


class Agent(BaseModel, Generic[TapeType]):
    """
    Base class for agents within the TapeAgents framework.

    An agent is a model that can run on a tape, generating new steps based on the tape's state.
    The agent can have subagents, which are other agents that it manages and can delegate to.
    The agent can also have nodes, which are atomic units of the agent's behavior that it can choose to run based on the tape.

    Attributes:
        name (str): The unique name of the agent.
        llms (dict[str, SerializeAsAny[LLM]]): A dictionary mapping names to LLM instances used by the agent.
        subagents (list[Any]): A list of subagents managed by this agent. Subagents must have unique names.
        templates (dict[str, Any]): A dictionary of templates used for generating prompts.
        nodes (list[SerializeAsAny[Node]]): A list of nodes that define the agent's actions and decision points.
        max_iterations (int): The maximum number of iterations the agent will execute before stopping.
        manager (Agent): Retrieves the manager agent overseeing this agent.
        llm (LLM): Default language model if only one is configured.
        template (Template): Default template if only one is configured.
        full_name (str): Hierarchical name of the agent, including its manager hierarchy.


    Raises:
        ValueError: If configuration inconsistencies are detected:

            - If a subagent is already managed by another agent
            - If any subagent is not an instance of Agent class
            - If there are duplicate names among subagents
            - If there are duplicate names among nodes

    """

    name: str = ""
    llms: dict[str, SerializeAsAny[LLM]] = {}
    # can't use list[Agent] because of pydantic bug
    # https://github.com/pydantic/pydantic/issues/9969
    subagents: list[Any] = Field(
        default_factory=lambda: [],
        description="List of subagents, which are agents that are run by this agent. Subagents must have unique names.",
    )
    templates: dict[str, Any] = {}
    nodes: list[SerializeAsAny[Node]] = Field(
        default_factory=lambda: [],
        description="List of nodes in the agent, order of the list used to determine the priority during activation. Nodes must have unique names.",
    )
    known_actions: list[type[Action] | ToolSpec] = Field(default_factory=list)
    tools_description: str = ""
    max_iterations: int = 100
    store_llm_calls: bool = False

    _manager: Any | None = None

    model_config = ConfigDict(extra="forbid")

    def model_post_init(self, __context: Any) -> None:
        if not self.name:
            # by default use the class name without the values for type variables
            # e.g. "Agent" instead of "Agent[Tape[...]]""
            self.name = self.__class__.__name__.split("[")[0]
        names = set()
        for i, agent in enumerate(self.subagents):
            if isinstance(agent, Agent):
                if agent._manager is not None:
                    raise ValueError("Agent is already a subagent of another agent. Make a copy of your agent.")
                agent._manager = self
            else:
                raise ValueError("Subagents must be instances of Agent")
            if agent.name in names:
                raise ValueError(
                    f'Duplicate subagent name "{agent.name}" in subagent {i}, pass a unique name to the subagent during creation'
                )
            names.add(agent.name)
        node_names = set()
        for i, node in enumerate(self.nodes):
            if node.name in node_names:
                raise ValueError(
                    f'Duplicate node name "{node.name}" in node {i}, pass a unique name to the node during creation'
                )
            if self.llms and node.llm not in self.llms:
                raise ValueError(
                    f"Node {node.name} references unknown LLM {node.llm}. Known LLMs: {list(self.llms.keys())}"
                )
            node_names.add(node.name)
        self.update_subagents()
        return super().model_post_init(__context)

    def update_subagents(self):
        for subagent in self.subagents:
            subagent.tools_description = self.tools_description
            subagent.known_actions = self.known_actions
            subagent.update_subagents()

    @property
    def manager(self):
        """
        Gets the manager of the agent.

        Returns:
            (Agent): The manager agent instance.

        Raises:
            ValueError: If the agent doesn't have a manager assigned.
        """
        if self._manager is None:
            raise ValueError("Agent doesn't have a manager")
        return self._manager

    @property
    def llm(self):
        """
        Get the default LLM instance associated with the agent.

        Returns:
            (LLM): The default LLM instance if only one LLM is configured.

        Raises:
            ValueError: If multiple LLMs are configured for this agent. In this case, use the `llms`
                       property to access specific LLM instances.
        """
        if len(self.llms) > 1:
            raise ValueError("Agent has multiple LLMs. Use llms property to access a specific one.")
        return self.llms[DEFAULT]

    @property
    def template(self):
        """
        Returns the default template of the agent.

        This property provides access to the default template when the agent has exactly one template.
        If multiple templates exist, it raises a ValueError indicating that specific templates
        should be accessed through the templates property instead.

        Returns:
            (Template): The default template object.

        Raises:
            ValueError: If the agent has more than one template.
            IndexError: If no templates exist (implicitly through list access).
        """
        if len(self.templates) > 1:
            raise ValueError("Agent has multiple templates. Use templates property to access a specific one.")
        return self.templates[DEFAULT]

    @property
    def full_name(self):
        """Returns the full hierarchical name of the agent.

        The full name is constructed by combining the manager's full name (if present)
        with this agent's name, separated by a forward slash. If the agent has no
        manager, returns just the agent's name.

        Returns:
            (str): The full hierarchical name path of the agent. Examples: "agent_name" (no manager), "manager_name/agent_name" (with manager)
        """
        if self._manager is None:
            return self.name
        return f"{self._manager.full_name}/{self.name}"

    def find_subagent(self, name: str):
        """
        Find a subagent by name in the list of subagents.

        Args:
            name (str): The name of the subagent to find.

        Returns:
            (Agent): The found subagent instance.

        Raises:
            ValueError: If no subagent with the given name is found.
        """
        for agent in self.subagents:
            if agent.name == name:
                return agent
        raise ValueError(f"Agent {name} not found")

    def find_node(self, name: str):
        """Find a node by its name in the list of nodes.

        Args:
            name (str): The name of the node to find.

        Returns:
            (Node): The node with the matching name.

        Raises:
            ValueError: If no node with the given name is found.
        """
        for node in self.nodes:
            if node.name == name:
                return node
        raise ValueError(f"Node {name} not found")

    def get_subagent_names(self) -> list[str]:
        """
        Returns a list of names of all subagents.

        Returns:
            list[str]: A list containing the names of all subagents in the agent.
        """
        return [agent.name for agent in self.subagents]

    def clone(self) -> Self:
        """
        Creates a deep copy of the current agent instance.

        This method creates an independent copy of the agent with all its attributes,
        but detaches it from any manager.

        Returns:
            Self: A new instance of the agent with identical attributes but no manager.
        """
        result = self.model_copy(deep=True)
        result._manager = None
        return result

    @classmethod
    def create(
        cls, llms: dict[str, LLM] | LLM | None = None, templates: dict[str, Any] | str | None = None, **kwargs
    ) -> Self:
        """
        Creates an instance of the class with provided LLMs and templates.

        Args:
            llms (Union[Dict[str, LLM], LLM, None]): Language model(s) to use. Can be:

                - A dictionary mapping names to LLM instances
                - A single LLM instance (will be mapped to default name)
                - None (empty dict will be used)
            templates (Union[Dict[str, Any], str, None]): Template(s) to use. Can be:

                - A dictionary mapping names to template configurations
                - A single template string (will be mapped to default name)
                - None (no templates will be used)
            **kwargs (dict, optional): Additional keyword arguments to pass to the class constructor

        Returns:
            Self: A new instance of the class initialized with the provided arguments

        Example:
            ```python
            agent = Agent.create(llm)  # Single LLM
            agent = Agent.create({"gpt": llm1, "claude": llm2})  # Multiple LLMs
            agent = Agent.create(llm, "template")  # LLM with template
            ```
        """
        if isinstance(llms, LLM):
            llms = {DEFAULT: llms}
        if isinstance(templates, str):
            templates = {DEFAULT: templates}
        if templates:
            kwargs["templates"] = templates

        return cls(llms=llms or {}, **kwargs)

    def update(self, agent_config: dict[str, Any]) -> Agent[TapeType]:
        """
        Updates the agent's configuration while preserving instance types.

        This method allows reconfiguration of the agent while maintaining the class types
        of LLMs and subagents. It performs a deep update by recursively applying changes
        to nested components.

        Args:
            agent_config (dict[str, Any]): New configuration dictionary containing LLMs,
                subagents, templates and other agent settings.

        Returns:
            Agent[TapeType]: A new agent instance with updated configuration.

        Raises:
            ValueError: If the new configuration has different LLMs or number of subagents
                than the current agent.

        Note:
            - Only string templates are updated, complex template objects are preserved
            - Node configurations are preserved to avoid potential issues
        """

        if not set(self.llms.keys()) == set(agent_config["llms"].keys()):
            raise ValueError("Agent has different LLMs than the new configuration.")
        if len(self.subagents) != len(agent_config["subagents"]):
            raise ValueError("Agent has different number of subagents than the new configuration.")
        # recurse into subagents
        subagents = [
            subagent.model_validate(subagent.update(subagent_obj))
            for subagent, subagent_obj in zip(self.subagents, agent_config["subagents"])
        ]
        # recurse into llms
        llms = {name: llm.model_validate(agent_config["llms"][name]) for name, llm in self.llms.items()}
        # only update templates are str
        templates = {
            name: (value if isinstance(value, str) else self.templates[name])
            for name, value in agent_config["templates"].items()
        }
        config_copy = agent_config.copy()
        config_copy["llms"] = llms
        config_copy["subagents"] = subagents
        config_copy["templates"] = templates
        # do not update nodes for now to avoid tricky bugs
        config_copy["nodes"] = self.nodes
        return type(self).model_validate(config_copy)

    def compute_view(self, tape: TapeType) -> TapeViewStack:
        """
        Compute the view stack from a given tape.

        Args:
            tape (TapeType): The input tape to process.

        Returns:
            TapeViewStack: A stack of views computed from the input tape.
        """
        return TapeViewStack.compute(tape)

    def select_node(self, tape: TapeType) -> Node:
        """
        Select the next node to execute based on the current state of the tape.

        The selection process follows these rules:
            1. If next_node is explicitly set in the tape view, return that node
            2. If no nodes have been run yet (last_node is None), return the first node
            3. Return the node that follows the last executed node in the list

        Args:
            tape (TapeType): The tape containing execution state and data

        Returns:
            Node: The next node to be executed

        Raises:
            ValueError: If unable to determine the next node to execute (e.g., reached end of list)
        """
        # Select the node to run next based on the current state of the tape.
        view = self.compute_view(tape).top
        if view.next_node:
            logger.debug(f"{self.name}: Next node was set explicitly in the tape: {view.next_node}")
            return self.find_node(view.next_node)

        if not view.last_node:
            logger.debug(f"{self.name}: No nodes have been run yet, select node 0: {self.nodes[0].name}")
            return self.nodes[0]

        # Select the next node that stored after the last node found in the tape
        logger.debug(f"{self.name}: Last node in view: {view.last_node}")
        logger.debug(f"{self.name}: Known nodes: {[node.name for node in self.nodes]}")
        for i, node in enumerate(self.nodes):
            if node.name == view.last_node and i + 1 < len(self.nodes):
                logger.debug(f"{self.name}: Select immediate next node: {self.nodes[i + 1].name}")
                return self.nodes[i + 1]
        raise ValueError("Next node not found")

    def make_prompt(self, tape: TapeType) -> Prompt:
        """
        Makes the prompt for the next iteration of the agent.
        This method generates a prompt by delegating to the selected node's make_prompt method.
        Can return a prompt with no messages, indicating the agent should generate next steps
        by following rules without LLM assistance. Agents that only delegate to subagents may
        not need to implement this method.

        Args:
            tape (TapeType): The tape containing the agent's state and history

        Returns:
            Prompt: A prompt object for the next agent iteration, potentially empty

        Note:
            - Empty prompts signal rule-based generation without LLM
            - Method may be optional for pure delegation agents
        """
        node = self.select_node(tape)
        prompt = node.make_prompt(self, tape)
        msg_debug = "\n\n".join([f"{m['role']}:\n{m.get('content')}" for m in prompt.messages])
        logger.debug(colored(f"Node {self.full_name}:{node.name}", "magenta"))
        logger.debug(colored(f"Prompt messages:\n{msg_debug}", "magenta"))
        logger.debug(colored(f"Prompt tools: {prompt.tools}", "magenta"))
        logger.debug(colored(f"Prompt response_format: {prompt.response_format}", "magenta"))
        return prompt

    def generate_steps(self, tape: TapeType, llm_stream: LLMStream) -> Generator[Step | PartialStep, None, None]:
        """
        Generate steps from the agent by selecting a node and processing its output.

        Args:
            tape (TapeType): The input tape containing the interaction history
            llm_stream (LLMStream): Stream interface for the language model output

        Yields:
            Union[Step, PartialStep]: Union[Step, PartialStep]: The generated steps or partial steps.
        """
        # Generate new steps and other events by feeding the prompt to the LLM
        node = self.select_node(tape)
        for step in node.generate_steps(self, tape, llm_stream):
            if isinstance(step, AgentStep):
                step.metadata.node = node.name
                if node.llm:
                    step.metadata.llm = node.llm
            yield step

    def make_llm_output(self, tape: TapeType, index: int) -> LLMOutput:
        """
        Generates an LLM output based on a tape and step index.

        Args:
            tape (TapeType): The input tape
            index (int): The position in the tape up to which to process.

        Returns:
            LLMOutput: The generated language model output for the tape segment.

        Note:
            This method delegates the actual output generation to the selected node's
            make_llm_output method after selecting the appropriate node based on the
            tape segment up to the given index.
        """
        return self.select_node(tape[:index]).make_llm_output(self, tape, index)

    def delegate(self, tape: TapeType) -> Agent[TapeType]:
        """
        Delegates control to the appropriate subagent based on the current tape state.

        This method recursively traverses the agent hierarchy to find the most specific
        subagent that should handle the current tape state based on views computed from
        the tape.

        Args:
            tape (TapeType): The tape containing the current state to process.

        Returns:
            Agent[TapeType]: The subagent that should handle the current tape state.
        """
        views = self.compute_view(tape)
        subagent = self
        for view in views.stack[1:]:
            subagent = subagent.find_subagent(view.agent_name)
        logger.debug(f"{self.full_name}: Delegating to subagent: {subagent.full_name}")
        return subagent

    def is_agent_step(self, step: Step) -> bool:
        """
        Check if a step was produced by the agent.

        Args:
            step (Step): The step object to check.

        Returns:
            bool: True if the step is an Action or Thought (agent-produced),
                  False otherwise.
        """
        return isinstance(step, (Action, Thought))

    def should_stop(self, tape: TapeType) -> bool:
        """
        Check if the agent should stop its turn and wait for observations.

        Args:
            tape (TapeType): The tape containing the sequence of steps (actions and observations).

        Returns:
            bool: True if the last step in the tape is an Action, indicating the agent should stop and wait for observations.
                 False if the last step is not an Action, indicating the agent can continue.
        """
        return isinstance(tape.steps[-1], Action)

    def run_iteration(
        self, tape: TapeType, llm_stream: LLMStream | None = None
    ) -> Generator[Step | PartialStep, None, None]:
        """
        Run one iteration of the agent (assuming one call to the underlyng model).

        During an iteration the agent generates steps from a stream of tokens that arises
        from a single LLM call with a single prompt. An agent can do multiple iterations
        before returning the next action (see `run` method).

        This function can also take a given `llm_stream`, which can be useful when the agent
        reuses a tape.

        Args:
            tape (TapeType): The tape to run the agent on
            llm_stream (LLMStream): The stream of tokens from the LLM

        Yields:
            Union[Step, PartialStep]: The generated steps or partial
        """
        if llm_stream is None:
            prompt = self.make_prompt(tape)
            if len(self.llms) > 1:
                llm = self.llms[self.select_node(tape).llm]
            elif len(self.llms) == 1:
                llm = self.llm
            llm_stream = llm.generate(prompt) if prompt else LLMStream(None, prompt)
        try:
            for step in self.generate_steps(tape, llm_stream):
                if isinstance(step, AgentStep):
                    step.metadata.prompt_id = llm_stream.prompt.id
                yield step
        except Exception as e:
            logger.exception(
                f" - Agent '{self.full_name}' - Node '{self.select_node(tape).name}' - step generation error: {e}"
            )
            raise e
        if self.store_llm_calls and (llm_call := getattr(llm_stream, "llm_call", None)):
            step.metadata.other["llm_call"] = llm_call

    def run(self, tape: TapeType, max_iterations: int | None = None) -> AgentStream[TapeType]:
        """
        Run the agent on the tape iteratively, delegating to subagents until a stop condition is met.

        This method executes the agent's logic by:
        1. Delegating to appropriate subagents based on the tape state
        2. Processing steps from subagent iterations
        3. Updating the tape with new steps
        4. Checking stop conditions
        5. Tracking metadata about the execution

        Args:
            tape (TapeType): The input tape to process
            max_iterations (int, optional): Maximum number of iterations to run.
                If None, uses self.max_iterations. Defaults to None.

        Returns:
            AgentStream[TapeType]: A stream of AgentEvents containing:

                - partial_step: Intermediate processing steps
                - step: Completed agent steps with updated tape
                - final_tape: Final tape with updated metadata after completion

        Yields:
            AgentEvent: Events indicating the agent's progress including partial steps,
                completed steps with updated tape, and the final result.

        Raises:
            ValueError: If the agent generates anything other than steps or partial steps.
        """
        if max_iterations is None:
            max_iterations = self.max_iterations

        def _run_implementation():
            nonlocal tape
            n_iterations = 0
            input_tape_length = len(tape)
            input_tape_id = tape.metadata.id
            stop = False
            original_metadata = tape.metadata
            while n_iterations < max_iterations and not stop:
                current_subagent = self.delegate(tape)
                for step in current_subagent.run_iteration(tape):
                    if isinstance(step, PartialStep):
                        yield AgentEvent(partial_step=step)
                    elif isinstance(step, AgentStep):
                        step.metadata.agent = current_subagent.full_name
                        tape = tape.append(step)
                        yield AgentEvent(step=step, partial_tape=tape)
                        if self.should_stop(tape):
                            stop = True
                    else:
                        raise ValueError("Agent can only generate steps or partial steps")
                n_iterations += 1
            updated_metadata = original_metadata.model_validate(
                dict(
                    parent_id=input_tape_id,
                    author=self.name,
                    n_added_steps=len(tape) - input_tape_length,
                )
            )
            final_tape = tape.model_copy(update=dict(metadata=updated_metadata))
            yield AgentEvent(final_tape=final_tape)

        return AgentStream(_run_implementation())

    def run_batch(self: Agent[TapeType], tapes: list[TapeType]) -> list[Tape]:
        """Run agent in parallel on tapes using batched LLM calls.

        This is faster than running agents in thread and having the LLM server batch the calls.

        """
        if len(self.llms) > 1:
            raise NotImplementedError("For run_agent_batch the agent must have only one LLM for now")
        if not isinstance(self.llm, TrainableLLM):
            raise NotImplementedError("For run_agent_batch the LLM must be TrainableLLM")
        original_tapes = list(tapes)
        n_iterations = 0
        active_indices = set(range(len(tapes)))
        while n_iterations < self.max_iterations:
            prompts = []
            current_subagents = [self.delegate(tapes[i]) for i in active_indices]
            prompts = [subagent.make_prompt(tape) for subagent, tape in zip(current_subagents, tapes)]
            llm_calls = self.llm.batch_generate(prompts)
            for i in active_indices:
                # Run the equivalent of agent.run_iteration
                llm_stream = LLMStream(
                    (LLMEvent(output=output) for output in (llm_calls[i].output,)), llm_calls[i].prompt
                )
                for step in self.generate_steps(tapes[i], llm_stream):
                    step.metadata.agent = current_subagents[i].full_name
                    if isinstance(step, AgentStep):
                        step.metadata.prompt_id = llm_calls[i].prompt.id
                    tapes[i] = tapes[i].append(step)
                    if self.should_stop(tapes[i]):
                        active_indices.remove(i)
                if self.store_llm_calls:
                    step.metadata.other["llm_call"] = llm_calls[i]
            n_iterations += 1
        for i in range(len(tapes)):
            updated_metadata = original_tapes[i].metadata.model_validate(
                dict(
                    parent_id=original_tapes[i].metadata.id,
                    author=self.name,
                    n_added_steps=len(tapes[i]) - len(original_tapes[i]),
                )
            )
            tapes[i] = tapes[i].model_copy(update=dict(metadata=updated_metadata))
        return tapes

    def reuse(self, tape: TapeType) -> tuple[TapeType, list[LLMCall]]:
        """
        Reuse another agent's tape as one's own.

        Construct LLM outputs at each step where a prompt is made. Check that output
        parsing yield the same steps as in the original tape. Rewrite metadata for all steps.

        Args:
            tape (TapeType): The tape to reuse

        Returns:
            tuple[TapeType, list[LLMCall]]: The reused tape and a list of LLM calls made during the reuse

        Raises:
            TapeReuseFailure: If the regenerated steps don't match the original tape.
        """
        reused_steps = []
        llm_calls = []
        i = 0
        while i < len(tape):
            past_tape = tape[:i]
            step = tape.steps[i]
            if self.is_agent_step(step):
                current_agent = self.delegate(past_tape)
                prompt = current_agent.make_prompt(past_tape)
                if not prompt:
                    reused_steps.append(step)
                    i += 1
                    continue
                output = current_agent.make_llm_output(tape, i)
                llm_call = LLMCall(prompt=prompt, output=output, cached=True)
                observe_llm_call(llm_call)

                # Validate that the reconstructed llm call leads to the same steps as in the given tape
                def _generator():
                    yield LLMEvent(output=output)

                new_steps = list(current_agent.run_iteration(past_tape, LLMStream(_generator(), prompt)))
                for j, new_step in enumerate(new_steps):
                    assert isinstance(new_step, Step)
                    old_step = tape.steps[i + j]
                    if type(old_step) is not type(new_step) or not _is_step_data_equal(old_step, new_step):
                        raise TapeReuseFailure(
                            f"Can't reuse tape because regenerated step {i + j} data doesn't match"
                            f"\nold step data: {old_step.llm_dict()}\nnew step data: {new_step.llm_dict()}",
                            partial_tape=past_tape,
                        )
                llm_calls.append(llm_call)
                reused_steps.extend(new_steps)
                i += len(new_steps)
            else:
                reused_steps.append(step)
                i += 1
        reused_tape = tape.model_validate(dict(context=tape.context, metadata=TapeMetadata(), steps=reused_steps))
        return reused_tape, llm_calls

    def get_node_runs(self, tape: TapeType) -> list[tuple[Node, int]]:
        """
        Parse the tape and identify the indices where each node began its execution.

        This method identifies transition points in the tape where different nodes started
        producing output by tracking changes in prompt IDs.

        Args:
            tape (TapeType): The sequence of tape steps to analyze.

        Returns:
            list[tuple[Node, int]]: List of tuples containing (node, index) pairs where:

                - node: The Node object that produced the tape fragment
                - index: The starting index in the tape where this node began execution
        """
        last_prompt_id = None
        result = []
        for index, step in enumerate(tape):
            if (prompt_id := step.metadata.prompt_id) and prompt_id != last_prompt_id:
                node = self.find_node(step.metadata.node)
                result.append((node, index))
            last_prompt_id = prompt_id
        return result

    def make_training_text(self, llm_call: LLMCall) -> TrainingText:
        """
        Routes the request to make training text to the agent's LLM.

        Args:
            llm_call (LLMCall): Object containing prompt and output from an LLM call.

        Returns:
            TrainingText: The training text generated from the prompt and output.

        Note:
            Currently only supports one LLM. Future versions will support multiple LLMs.
        """
        # TODO: support more than 1 LLM
        return self.llm.make_training_text(llm_call.prompt, llm_call.output)

    def make_training_data(self, tape: TapeType) -> list[TrainingText]:
        """
        Generates training data from a tape by converting LLM calls into training texts.

        Args:
            tape (TapeType): A tape containing recorded LLM interactions.

        Returns:
            list[TrainingText]: A list of training text objects created from the LLM calls.

        Notes:
            This method first reuses the tape to extract LLM calls, then converts each call
            into a training text format using make_training_text().
        """
        _, llm_calls = self.reuse(tape)
        return [self.make_training_text(llm_call) for llm_call in llm_calls]

    async def arun(self, tape: TapeType, session: aiohttp.ClientSession, max_iterations: int | None = None):
        if max_iterations is None:
            max_iterations = self.max_iterations

        n_iterations = 0
        input_tape_length = len(tape)
        input_tape_id = tape.metadata.id
        stop = False
        original_metadata = tape.metadata
        while n_iterations < max_iterations and not stop:
            current_subagent = self.delegate(tape)
            async for step in current_subagent.arun_iteration(tape, session):
                if isinstance(step, PartialStep):
                    yield AgentEvent(partial_step=step)
                elif isinstance(step, AgentStep):
                    step.metadata.agent = current_subagent.full_name
                    tape = tape.append(step)
                    yield AgentEvent(step=step, partial_tape=tape)
                    if self.should_stop(tape):
                        stop = True
                else:
                    raise ValueError("Agent can only generate steps or partial steps")
            n_iterations += 1
        updated_metadata = original_metadata.model_validate(
            dict(
                parent_id=input_tape_id,
                author=self.name,
                n_added_steps=len(tape) - input_tape_length,
            )
        )
        final_tape = tape.model_copy(update=dict(metadata=updated_metadata))
        yield AgentEvent(final_tape=final_tape)

    async def arun_iteration(self, tape: TapeType, session: aiohttp.ClientSession):
        if len(self.llms) > 1:
            llm = self.llms[self.select_node(tape).llm]
        else:
            llm = self.llm
        prompt = self.make_prompt(tape)
        t = time.perf_counter()
        llm_stream = await llm.agenerate(prompt, session) if prompt else LLMStream(None, prompt)
        llm_call_time = time.perf_counter() - t
        step = None
        time_saved = False
        for step in self.generate_steps(tape, llm_stream):
            if isinstance(step, AgentStep):
                step.metadata.prompt_id = llm_stream.prompt.id
                if not time_saved:
                    step.metadata.other["llm_call_time"] = llm_call_time
                    time_saved = True
            yield step
        if (
            step
            and not isinstance(step, PartialStep)
            and self.store_llm_calls
            and (llm_call := getattr(llm_stream, "llm_call", None))
        ):
            step.metadata.other["llm_call"] = llm_call
        if step is None:
            raise ValueError("No step was generated during the iteration.")


class TapeReuseFailure(ValueError):
    """Exception raised when tape reuse operation fails.

    This exception is raised when an attempt to reuse a tape encounters an error,
    providing access to the succesfully reused part of the tape

    Args:
        msg (str): Description of why the tape reuse failed
        partial_tape (Tape): The incomplete/partial tape that was being constructed

    Attributes:
        partial_tape (Tape): The incomplete tape at the point of failure
    """

    def __init__(self, msg: str, partial_tape: Tape):
        self.partial_tape = partial_tape
        super().__init__(msg)


def _is_step_data_equal(step1: Step, step2: Step) -> bool:
    """Compare steps ignoring the metadata.

    Some LLM API messages include metadata, like unique tool call ids, that
    we must store and send back to the API. This function excludes this metadata from the comparison.

    Another caveat is that some data, like function call arguments in LLM messages, is not deserialized,
    and hence can be tricky to compare across steps. This function deserializes known fields like that before comparison..

    Args:
        step1 (Step): The first step to compare
        step2 (Step): The second step to compare

    Returns:
        bool: True if the data in the steps is equal, False otherwise
    """

    def just_data(step: Step) -> dict:
        if isinstance(step, LLMOutputParsingFailureAction):
            return {}

        data = step.llm_dict()
        for tc in data.get("tool_calls", []):
            tc.pop("id", None)
            if function := tc.get("function"):
                function["arguments"] = json.loads(function["arguments"])
        data.pop("tool_call_id", None)
        return data

    return just_data(step1) == just_data(step2)


class Annotator(Agent[AnnotatorTapeType], Generic[TapeType, AnnotatorTapeType]):
    """
    Annotator is the base class for agents that produce annotations for the tape of another agent.
    It annotates the tape by converting it into its own tape and then producing an annotation step appended to the converted tape.
    """

    @abstractmethod
    def make_own_tape(self, tape: TapeType) -> AnnotatorTapeType:
        pass

    def annotate(self, tape: TapeType) -> AnnotatorTapeType:
        return self.run(self.make_own_tape(tape)).get_final_tape()

    def get_annotation(self, own_tape: AnnotatorTapeType) -> Any:
        return own_tape.steps[-1].annotation


class ObservationMaker(Agent[ObservationMakerTapeType], Generic[TapeType, ObservationMakerTapeType]):
    @abstractmethod
    def make_own_tape(self, tape: TapeType) -> ObservationMakerTapeType:
        pass

    def add_observation(self, tape: TapeType, own_tape: ObservationMakerTapeType) -> TapeType:
        last_step = own_tape.steps[-1]
        if not isinstance(own_tape.steps[-1], MakeObservation):
            raise ValueError("Own tape must end with MakeObservation to add observation to the target tape")
        metadata = TapeMetadata(
            parent_id=tape.metadata.id, author_tape_id=own_tape.metadata.id, author=self.name, n_added_steps=1
        )
        return tape.append(last_step.new_observation).model_copy(update=dict(metadata=metadata))

    def can_continue(self, _: TapeType) -> bool:
        return True

    def continue_tape(self, tape: TapeType) -> TapeType:
        own_tape = self.make_own_tape(tape)
        for event in self.run(own_tape):
            if event.final_tape:
                return self.add_observation(tape, event.final_tape)
        raise ValueError("Observation maker didn't produce final tape")



import json
import logging
import os
from itertools import zip_longest
from typing import Any, Callable

from pydantic import Field
from termcolor import colored

from tapeagents.config import DB_DEFAULT_FILENAME
from tapeagents.core import LLMOutput, Prompt, TrainingText
from tapeagents.llms.base import LLM, LLMCall, LLMEvent, LLMStream, closest_prompt
from tapeagents.observe import retrieve_all_llm_calls
from tapeagents.utils import FatalError, diff_strings

logger = logging.getLogger(__name__)


class ReplayLLM(LLM):
    """
    Specialized LLM class that replays previously recorded LLM interactions.

    Loads and replays model interactions from a SQLite database, allowing for
    deterministic replay of previous LLM conversations without making new API calls.

    The class is useful for:

    - Testing and debugging LLM interactions
    - Reproducing specific model behaviors
    - Avoiding repeated API calls during development
    - Creating deterministic test scenarios

    Attributes:
        outputs (dict[str, str]): Dictionary mapping prompt strings to their recorded outputs
        llm_calls (list[LLMCall]): List of recorded LLM call objects
        count_tokens_fn (Callable): Function to count tokens in prompts/messages
        make_training_text_fn (Callable): Function to create training text from prompt/output pairs

    Raises:
        FatalError: When a prompt is not found in the recorded outputs
        AssertionError: When the specified SQLite database file doesn't exist
    """

    outputs: dict[str, str] = Field(default_factory=dict)
    llm_calls: list[LLMCall]
    count_tokens_fn: Callable = lambda x: 0
    make_training_text_fn: Callable = lambda x, y: TrainingText(text="", n_predicted=0)
    _get_step_schema: Callable = lambda: None

    @classmethod
    def from_llm(cls, llm: LLM, run_dir: str, prompts_file: str = DB_DEFAULT_FILENAME):
        """
        Create a ReplayLLM instance from an existing LLM and a SQLite database file.

        Args:
            cls (Type): The class to instantiate.
            llm (LLM): The original LLM instance.
            run_dir (str): The directory where the SQLite database file is located.
            prompts_file (str, optional): The name of the SQLite database file. Defaults to DB_DEFAULT_FILENAME.

        Returns:
            (ReplayLLM): An instance of ReplayLLM initialized with the LLM calls from the SQLite database.

        Raises:
            AssertionError: If the SQLite database file does not exist at the specified path.
        """
        sqlite_fpath = os.path.join(run_dir, prompts_file)
        assert os.path.exists(sqlite_fpath), f"Sqlite not found: {sqlite_fpath}"
        llm_calls = retrieve_all_llm_calls(sqlite_fpath)
        replay_llm = ReplayLLM(
            llm_calls=llm_calls,
            model_name=llm.tokenizer_name or llm.model_name,
            context_size=llm.context_size,
        )
        replay_llm.tokenizer = llm.tokenizer
        replay_llm.count_tokens_fn = llm.count_tokens
        replay_llm.make_training_text_fn = llm.make_training_text
        replay_llm._get_step_schema = llm.get_step_schema
        return replay_llm

    def get_step_schema(self, *args, **kwargs):
        return self._get_step_schema(*args, **kwargs)

    def model_post_init(self, __context: Any) -> None:
        dups = 0
        for llm_call in self.llm_calls:
            prompt_key = json.dumps(llm_call.prompt.messages, indent=2, ensure_ascii=False, sort_keys=True)
            output = llm_call.output.content or ""
            if prompt_key in self.outputs and output != self.outputs[prompt_key]:
                logger.debug(f"Output duplicate, using last value!\nOLD:{self.outputs[prompt_key]}\nNEW:{output}")
                dups += 1
            self.outputs[prompt_key] = output
        logger.info(f"Loaded {len(self.outputs)} outputs, {dups} duplicates")
        return super().model_post_init(__context)

    def generate(self, prompt: Prompt, **kwargs) -> LLMStream:
        """
        Generates an LLMStream based on the provided prompt.

        This method checks if the prompt has been previously processed and cached. If a cached output is found,
        it is returned. Otherwise, it attempts to find the closest known prompt and logs the differences. If no
        similar prompt is found, a FatalError is raised.

        Args:
            prompt (Prompt): The prompt object containing the messages to be processed.
            **kwargs (dict, optional): Additional keyword arguments.

        Returns:
            LLMStream: A stream of LLM events containing the generated output.

        Raises:
            FatalError: If the prompt is not found in the cache and no similar prompt is found.
        """

        def _implementation():
            prompt_key = json.dumps(prompt.messages, indent=2, ensure_ascii=False, sort_keys=True)
            if prompt_key in self.outputs:
                logger.debug(colored("prompt cache hit", "green"))
                output = self.outputs[prompt_key]
            elif len(prompt_key) < 20000:
                logger.warning(
                    colored(f"prompt of size {len(prompt_key)} not found, checking similar ones..", "yellow")
                )
                known_prompts = list(self.outputs.keys())
                closest, score = closest_prompt(prompt_key, known_prompts)
                if score >= 0.7:
                    logger.warning(f"Closest prompt score {score:.3f}")
                    for i, (a, b) in enumerate(zip_longest(prompt.messages, json.loads(closest), fillvalue={})):
                        aa = a.get("content", str(a))
                        bb = b.get("content", str(b))
                        if aa == bb:
                            continue
                        if len(aa) < 300 and len(bb) < 300:
                            logger.warning(f"STEP{i} A:\n{aa}\nSTEP{i} B:\n{bb}")
                        else:
                            logger.warning(f"STEP{i}: {diff_strings(aa, bb)}\n")
                raise FatalError("prompt not found")
            else:
                messages_previews = []
                for m in prompt.messages:
                    try:
                        if isinstance(m["content"], list):
                            msg = "[text,img]"
                        else:
                            m_dict = json.loads(m["content"])
                            msg = list(m_dict.keys())
                        messages_previews.append({"role": m["role"], "content": msg})
                    except Exception:
                        messages_previews.append({"role": m["role"], "content": "text"})
                logger.warning(
                    f"prompt with {len(prompt.messages)} messages {messages_previews}, {len(prompt_key)} chars not found, skipping.."
                )
                raise FatalError("prompt not found")
            yield LLMEvent(output=LLMOutput(content=output))

        return LLMStream(_implementation(), prompt=prompt)

    def make_training_text(self, prompt: Prompt, output: LLMOutput) -> TrainingText:
        """
        Generates training text based on the provided prompt and output.

        Args:
            prompt (Prompt): The input prompt to generate training text from.
            output (LLMOutput): The output generated by the language model.

        Returns:
            TrainingText: The generated training text.
        """
        return self.make_training_text_fn(prompt, output)

    def count_tokens(self, messages: list[dict] | str) -> int:
        """
        Counts the number of tokens in the given messages.

        Args:
            messages (Union[list[dict], str]): A list of message dictionaries or a single string message.

        Returns:
            int: The total number of tokens in the messages.
        """
        return self.count_tokens_fn(messages)



# Add lifespan support for startup/shutdown with strong typing
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from mcp.server.fastmcp import Context, FastMCP

from tapeagents.tools.simple_browser import SimpleTextBrowser


@dataclass
class AppContext:
    browser: SimpleTextBrowser


@asynccontextmanager
async def app_lifespan(server: FastMCP) -> AsyncIterator[AppContext]:
    """Manage application lifecycle with type-safe context"""
    # Initialize on startup
    browser = SimpleTextBrowser()
    try:
        yield AppContext(browser=browser)
    finally:
        # Save cache on shutdown
        browser.flush_cache()


# Pass lifespan to server
mcp = FastMCP("Simple Browser Server", lifespan=app_lifespan)


# Access type-safe lifespan context in tools
@mcp.tool()
def get_page(url: str, ctx: Context) -> str:
    """Tool that returns the text of the document at the given URL"""
    browser: SimpleTextBrowser = ctx.request_context.lifespan_context.browser
    text, pages, error = browser.get_page(url)
    if error:
        raise ValueError(f"Error fetching document: {error}")
    return text


@mcp.tool()
def echo_tool(message: str) -> str:
    """Echo a message as a tool"""
    return f"Tool echo: {message}"


if __name__ == "__main__":
    mcp.run()



"""
Views and view stacks for subagents context isolation.
"""

from __future__ import annotations

from collections import defaultdict
from typing import Generic, Literal

from pydantic import BaseModel, Field

from tapeagents.core import AgentStep, Call, ControlFlow, Observation, Respond, SetNextNode, StepType, Tape, Thought


class Broadcast(Thought):
    """
    Broadcast a message to many subagents.

    The current agent remains active.

    Attributes:
        content (str): The content of the message to broadcast.
        from_ (str): The name of the agent broadcasting the message.
        to (list[str]): The list of subagents to broadcast the message to.
        kind (Literal["broadcast"]): The kind of the step, which is "broadcast".
    """

    content: str
    from_: str
    to: list[str]
    kind: Literal["broadcast"] = "broadcast"


class TapeView(BaseModel, Generic[StepType]):
    """
    Ephemeral view of an agent's part of the tape.

    Presents tape data in the form that is describing for describing the agent's logic.

    Attributes:
        agent_name (str): The name of the agent.
        agent_full_name (str): The full name of the agent, including the names of all its parent agents.
        steps (list[StepType]): The list of steps in the agent's part of the tape.
        steps_by_kind (dict[str, list[StepType]]): A dictionary of steps grouped by their kind.
        outputs_by_subagent (dict[str, StepType]): A dictionary of the output steps of the agent's subagents.
        last_node (str): The last node the agent was in.
        last_prompt_id (str): The ID of the prompt of the last step.
        next_node (str): The next node the agent will go to next.
    """

    agent_name: str
    agent_full_name: str
    steps: list[StepType] = []
    steps_by_kind: dict[str, list[StepType]] = {}
    outputs_by_subagent: dict[str, StepType] = {}
    last_node: str = ""
    last_prompt_id: str = ""
    next_node: str = ""

    def add_step(self, step: StepType):
        self.steps.append(step)
        kind = step.kind  # type: ignore
        if kind not in self.steps_by_kind:
            self.steps_by_kind[kind] = []
        if self.is_step_by_active_agent(step):
            if step.metadata.prompt_id != self.last_prompt_id and not isinstance(step, Respond):
                # respond should not reset the previous set_next_node in the caller agent
                self.next_node = ""
            self.last_prompt_id = step.metadata.prompt_id
            self.last_node = step.metadata.node
            if isinstance(step, SetNextNode):
                self.next_node = step.next_node
        self.steps_by_kind[kind].append(step)

    def get_output(self, subagent_name_or_index: int | str) -> StepType:
        if isinstance(subagent_name_or_index, int):
            return list(self.outputs_by_subagent.values())[subagent_name_or_index]
        return self.outputs_by_subagent[subagent_name_or_index]

    def is_step_by_active_agent(self, step: StepType):
        # state machine doesn't know the name of the root agent, so in the comparison here
        # we need cut of the first component
        if not isinstance(step, AgentStep):
            return False
        parts_by = step.metadata.agent.split("/")
        parts_frame_by = self.agent_full_name.split("/")
        return parts_by[1:] == parts_frame_by[1:]


class TapeViewStack(BaseModel, Generic[StepType]):
    """
    Stack of tape views of the agents in the call chain.

    If Agent A calls Agent B, and then Agent B calls Agent C,
    the stack will looks as follows:
    0: TapeView of Agent A
    1: TapeView of Agent B
    2: TapeView of Agent C

    Attributes:
        stack (list[TapeView[StepType]]): The stack of tape views.
        messages_by_agent (dict[str, list[Union[Call, Respond, Broadcast]]]): A dictionary of messages by agent.
    """

    stack: list[TapeView[StepType]]
    messages_by_agent: dict[str, list[Call | Respond | Broadcast]] = Field(default_factory=lambda: defaultdict(list))

    @property
    def top(self):
        return self.stack[-1]

    def update(self, step: StepType):
        """
        Updates the view based on the given step type.

        Args:
            step (StepType): The step to process. Can be one of:

                - Call: Creates and pushes new view onto the stack
                - Broadcast: Processes broadcast messages
                - Respond: Removes top view from stack
                - AgentStep: Adds step to current top view
                - Observation: Adds observation to current top view

        Raises:
            ValueError: If the step type is not supported
        """
        top = self.stack[-1]
        match step:
            case Call():
                self.put_new_view_on_stack(step)
            case Broadcast():
                self.broadcast(step)
            case Respond():
                self.pop_view_from_stack(step)
            case AgentStep():
                top.add_step(step)
            case Observation():
                top.add_step(step)
            case _:
                raise ValueError(f"Unsupported step type {type(step)}")

    def pop_view_from_stack(self, step: Respond):
        top = self.stack[-1]
        self.stack.pop()
        new_top = self.stack[-1]

        if step.copy_output:
            for top_step in reversed(top.steps):
                # How we choose the output step of the frame.
                # - exclude the input steps that the caller agent added to the tape for the given agent
                #   (note that by this line the former caller agent is the active agent)
                # - exclude Call and Respond steps
                # - exclude Observation steps
                # - among the remaining steps pick the last one
                if not self.top.is_step_by_active_agent(top_step) and not isinstance(
                    top_step, (ControlFlow, Observation)
                ):
                    new_top.add_step(top_step)
                    new_top.outputs_by_subagent[top.agent_name] = top_step
                    break
                    # TODO: what if the agent was not called by its immediate manager?

        receiver = step.metadata.agent.rsplit("/", 1)[0]
        self.messages_by_agent[step.metadata.agent].append(step)
        self.messages_by_agent[receiver].append(step)
        new_top.add_step(step)

    def broadcast(self, step):
        top = self.stack[-1]
        top.add_step(step)
        for to in step.to:
            receiver = f"{step.metadata.agent}/{to}"
            self.messages_by_agent[receiver].append(step)

    def put_new_view_on_stack(self, step):
        top = self.stack[-1]
        top.add_step(step)
        self.stack.append(
            TapeView(
                agent_name=step.agent_name,
                agent_full_name=top.agent_full_name + "/" + step.agent_name,
            )
        )
        self.stack[-1].add_step(step)
        receiver = f"{step.metadata.agent}/{step.agent_name}"
        self.messages_by_agent[step.metadata.agent].append(step)
        self.messages_by_agent[receiver].append(step)

    @staticmethod
    def compute(tape: Tape, root_agent_name: str = "root") -> TapeViewStack[StepType]:
        """
        Computes a stack of tape views from a given tape.

        This function processes a tape step by step and builds a view stack that tracks
        the execution flow through different agents.

        Args:
            tape (Tape): The tape containing steps to process
            root_agent_name (str, optional): Name of the root agent. Defaults to "root"

        Returns:
            TapeViewStack[StepType]: A stack of tape views representing the execution flow

        Note:
            TODO: Implement view retrieval from tape prefix and recomputation functionality
        """
        stack = TapeViewStack(stack=[TapeView(agent_name=root_agent_name, agent_full_name=root_agent_name)])
        for step in tape.steps:
            stack.update(step)
        return stack  # type: ignore



"""Generate the code reference pages."""

from pathlib import Path

import mkdocs_gen_files

project = "tapeagents"
nav = mkdocs_gen_files.Nav()
root = Path(__file__).parent.parent
src = root / project

for path in sorted(src.rglob("*.py")):
    module_path = path.relative_to(src).with_suffix("")
    doc_path = path.relative_to(src).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = (project,) + module_path.parts

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1] == "__main__":
        continue
    if not doc_path:
        continue

    nav_aliases = {
        "io": "IO",
        "rl": "RL",
        "llms": "LLMs",
        "llm_function": "LLM Function",
    }
    nav_key = list(parts[1:])
    if nav_key:
        for i in range(len(nav_key)):
            if nav_key[i] in nav_aliases:
                nav_key[i] = nav_aliases[nav_key[i]]
            else:
                nav_key[i] = nav_key[i].title().replace("_", " ").strip()
        nav_key = tuple(nav_key)
        nav[nav_key] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        identifier = ".".join(parts)
        if identifier:
            print(f"::: {identifier}", file=fd)

    mkdocs_gen_files.set_edit_path(full_doc_path, path.relative_to(root))

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



import logging
from typing import Any, Generator, Literal, Union

from pydantic import Field

from tapeagents.agent import Agent
from tapeagents.core import LLMOutputParsingFailureAction, Observation, Prompt, Step, Tape, Thought
from tapeagents.llms import LLM
from tapeagents.nodes import StandardNode

logger = logging.getLogger(__name__)


class Task(Observation):
    kind: Literal["task"] = "task"
    task: str
    template: str = Field(
        description="Template for the task. Should contain a {task} placeholder for the task text.", default="{task}"
    )

    def llm_view(self, indent: int | None = 2) -> str:
        return self.template.format(task=self.task)


class ReasoningThought(Thought):
    """
    Thoughts produced by the agent during the reasoning process.
    """

    kind: Literal["reasoning_thought_with_value"] = "reasoning_thought_with_value"
    reasoning: str = Field(description="chain of thoughts")


RLMathTape = Tape[
    None,
    Union[
        Task,
        ReasoningThought,
        LLMOutputParsingFailureAction,
    ],
]


class ReasoningNode(StandardNode):
    max_prompt_length: int = 1024

    def parse_completion(self, completion: str) -> Generator[Step, None, None]:
        try:
            step = ReasoningThought(reasoning=completion)
        except Exception as e:
            logger.info(f"Failed to parse agent output: {completion}\n\nError: {e}")
            yield LLMOutputParsingFailureAction(
                error=f"Failed to parse agent output: {completion}\n\nError: {e}", llm_output=completion
            )
            return
        yield step

    def make_prompt(self, agent: Any, tape: Tape) -> Prompt:
        messages = []
        if self.system_prompt:
            messages.append({"role": "system", "content": self.system_prompt})

        # the tape is only step long and it is the task
        task = tape.steps[0]
        assert isinstance(task, Task), f"Expected a Task, got {task.__class__.__name__}"
        messages.append({"role": "user", "content": task.llm_view()})
        prompt_token_ids = agent.llm.tokenizer.apply_chat_template(
            messages, add_special_tokens=True, add_generation_prompt=True
        )
        prompt_token_ids = prompt_token_ids[-self.max_prompt_length :]
        return Prompt(messages=messages, token_ids=prompt_token_ids)


#### Agent and Environment ####
class CoTMathAgent(Agent):
    @classmethod
    def create(cls, system_prompt: str, llm: LLM, max_prompt_length: int):
        agent = super().create(
            llm,
            nodes=[
                ReasoningNode(
                    name="cot",
                    steps=ReasoningThought,
                    system_prompt=system_prompt,  # if system_prompt else "",
                    max_prompt_length=max_prompt_length,
                ),
            ],
            max_iterations=1,
        )
        agent.store_llm_calls = True
        if agent.llm.tokenizer is None:
            agent.llm.load_tokenizer()
        return agent



import copy
import logging
import multiprocessing
import os
import random
import re
import time
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from itertools import chain
from pathlib import Path
from typing import Dict, List, Tuple

import hydra
import numpy as np
import torch
import wandb
from datasets import load_dataset
from omegaconf import DictConfig, OmegaConf
from termcolor import colored
from tqdm import tqdm

from tapeagents.agent import Agent
from tapeagents.core import StepMetadata, TrainingText
from tapeagents.finetune.data import MASKED_TOKEN_ID
from tapeagents.finetune.logging_ import flatten_dict_config, init_wandb
from tapeagents.llms import TrainableLLM

from .cot_math_agent import CoTMathAgent, RLMathTape, Task
from .utils import VLLMServiceManager, calculate_stats, clean_up, launch_training, load_state, save_state, setup_logging

logger = logging.getLogger(__name__)


def process_gsm8k(item):
    answer = item["answer"].split("#### ")[-1]
    sample = {"dataset": "gsm8k-cot", "task": item["question"], "answer": answer}
    return sample


def load_datasets() -> Tuple[list, list]:
    train_dataset = load_dataset("openai/gsm8k", "main", split="train", trust_remote_code=True)
    test_dataset = load_dataset("openai/gsm8k", "main", split="test", trust_remote_code=True)
    train_samples = [process_gsm8k(s) for s in tqdm(train_dataset, desc="Processing train samples")]
    test_samples = [process_gsm8k(s) for s in tqdm(test_dataset, desc="Processing test samples")]
    logger.info(f"Loaded {len(train_samples)} training samples")
    logger.info(f"Loaded {len(test_samples)} test samples")
    return train_samples, test_samples


def batch_annotate_traces_with_ref_logprobs(llm: TrainableLLM, traces: List[TrainingText]):
    """
    Annotates training traces with reference model log probabilities.

    Args:
        llm (TrainableLLM): The reference language model to get log probabilities from
        traces (List[TrainingText]): List of training traces to annotate

    Returns:
        None: The traces are modified in-place by adding ref_logprobs
    """
    prompt_token_ids = []
    completion_token_ids = []
    for trace in traces:
        prompt_token_ids.append(trace.input_ids[: -len(trace.logprobs)])
        completion_token_ids.append(trace.input_ids[-len(trace.logprobs) :])
    try:
        all_ref_logprobs = llm.get_batch_logprobs_token_ids(prompt_token_ids, completion_token_ids)
    except Exception as e:
        logger.error(f"Failed to get ref logprobs: {e}")
        return
    for trace, ref_logprobs in zip(traces, all_ref_logprobs):
        trace.ref_logprobs = [c["logprob"] for c in ref_logprobs["content"]]
        assert len(trace.ref_logprobs) == len(trace.logprobs), f"{len(trace.ref_logprobs)} != {len(trace.logprobs)}"


def convert_problems_to_tapes(problems: list, cfg: DictConfig) -> list[RLMathTape]:
    """
    Creates RLMathTape objects from a list of math problem dictionaries.

    Args:
        problems (list[dict]): List of dictionaries containing math problems, where each dict
            has 'question' and expected answer value. The list is created from a dataset.

    Returns:
        list[RLMathTape]: List of RLMathTape objects initialized with the math problems as Task steps.
    """
    tapes: list[RLMathTape] = []
    for problem in problems:
        start_step = Task(
            task=problem["task"],
            template=cfg.task_template,
            metadata=StepMetadata(
                other={
                    "value": problem["answer"],  # expected answer
                }
            ),
        )
        tape = RLMathTape(steps=[start_step], context=None)
        tapes.append(tape)
    return tapes


def extract_tape_training_samples(
    new_tape: RLMathTape, agent: CoTMathAgent, cfg: DictConfig
) -> Tuple[List[TrainingText], Dict[str, int]]:
    """
    Process a single tape to extract training samples and statistics.

    Args:
        new_tape: The tape to process containing math problem steps
        agent: CoTMathAgent
        tapes_dir: Directory to save processed tapes
        cfg: Configuration

    Returns:
        Tuple[List[TrainingText], Dict[str, int]]:
            - List of training samples (`TrainingText`) extracted from the tape.
            - Dictionary with tape-level statistics, including:
                - 'reward': Reward for the tape (-1, 0, or 1).
                - 'success': 1 if the answer is correct, 0 otherwise.
                - 'no_error': 1 if the LLM output is parsable, 0 otherwise.
                - 'prompt_tokens': Total prompt tokens used in the tape.
                - 'output_tokens': Total output tokens generated in the tape.
                - 'overflows': Number of times the output overflowed token limit.
    """
    tape_prompt_tokens = 0
    tape_output_tokens = 0

    reasoning = new_tape.steps[-1].reasoning
    begin_boxed = reasoning.rfind("\\boxed")
    prediction = reasoning[begin_boxed:]
    prediction = prediction.replace("\\boxed{", "").replace("}", "").strip()
    number_match = re.search(r"-?\d+", prediction)
    if not number_match:
        # LLM produced a step that was unparsable. Negative reward.
        no_error, reward, success = 0, -1, 0
    else:
        # LLM did respect the formatting
        no_error = 1
        prediction = number_match.group()
        answer = new_tape.steps[0].metadata.other["value"]

        if prediction == answer:
            # Correct answer
            reward, success = 1, 1
        else:
            # Incorrect answer or no answer
            reward, success = 0, 0

    training_samples: list[TrainingText] = []
    # For each LLM interaction in the tape:
    # - Create a training sample from the prompt and output
    # - Get log probabilities of the output tokens
    # - Set group ID for tracking and advantage calculation
    overflows = []
    max_num_tokens = agent.llm.parameters["max_tokens"]
    for step in new_tape.steps:
        if llm_call := step.metadata.other.get("llm_call"):
            tape_prompt_tokens += llm_call.prompt_length_tokens
            tape_output_tokens += llm_call.output_length_tokens
            overflow = True if llm_call.output_length_tokens == max_num_tokens else False
            overflows.append(overflow)

            if logprobs := getattr(llm_call, "logprobs", None):
                trace = agent.llm.make_training_text(llm_call.prompt, llm_call.output)
                prompt_token_ids = agent.llm.tokenizer.apply_chat_template(
                    llm_call.prompt.messages, add_special_tokens=True, add_generation_prompt=True
                )

                input_ids = prompt_token_ids + [lp.token_id for lp in logprobs]
                labels = [lp.token_id for lp in logprobs]
                # MASKED_TOKEN_ID is -100 and is the default "ignore_index" in nn.CrossEntropyLoss,
                # see https://pytorch.org/docs/stable/generated/torch.nn.CrossEntropyLoss.html
                labels = [MASKED_TOKEN_ID] * (len(input_ids) - len(labels)) + labels

                trace.input_ids = input_ids
                trace.labels = labels

                reward = -1 if overflow else reward
                trace.logprobs = [lp.logprob for lp in llm_call.logprobs]
                trace.group_id = new_tape.metadata.parent_id
                trace.reward = reward
                training_samples.append(trace)

    tape_stats = {
        "reward": reward,
        "success": success,
        "no_error": no_error,
        "prompt_tokens": tape_prompt_tokens,
        "output_tokens": tape_output_tokens,
        "overflows": sum(overflows),
    }
    return training_samples, tape_stats


def batch_run_agent_replica(agent: CoTMathAgent, tapes: list[RLMathTape]) -> tuple[Agent, list[RLMathTape]]:
    """
    Returns:
    tuple[Agent, list[RLMathTape]]: A tuple containing:
        - agent (Agent): The updated agent after running the batch of tapes.
                         Agent state might be modified during the `run_batch` operation.
        - final_tapes (list[RLMathTape]): The list of `RLMathTape` objects after the agent
                                         has run on them. These tapes will be completed with agent interactions.
    """
    final_tapes = agent.run_batch(tapes)
    # There is some statistics that we track in the agent in a mutable way
    return agent, final_tapes


def generate_training_data(
    agent_replicas: list[CoTMathAgent],
    tapes: list[RLMathTape],
    cfg: DictConfig,
    tapes_dir: Path,
    split_name: str,
) -> Tuple[list[CoTMathAgent], List[RLMathTape], List[TrainingText], Dict[str, float]]:
    """
    Generate complete tapes and training samples from a list of initialized tapes.

    Args:
        agent: Agent that interacts with the math environment
        tapes: List of tapes initialized with math problems
        cfg: Configuration
        tapes_dir: Directory to save processed episodes
        split_name: Name of split ('train' or other)

    Returns:
        Tuple containing:
        - List of completed RLMathTapes
        - List of training samples with rewards and logprobs
        - Dictionary of performance statistics and execution times
    """

    start_make_data = time.time()
    os.makedirs(tapes_dir, exist_ok=True)
    reward_stats = defaultdict(list)
    no_errors_stats = defaultdict(list)
    success_stats = defaultdict(list)
    prompt_tokens_stats = defaultdict(list)
    output_tokens_stats = defaultdict(list)
    overflow_stats = defaultdict(list)
    training_samples: List[TrainingText] = []

    logger.info(f"Run the agent on {split_name}")

    start_making_tapes = time.time()
    with ProcessPoolExecutor(max_workers=len(agent_replicas)) as executor:
        replica_tapes = [tapes[i :: len(agent_replicas)] for i in range(len(agent_replicas))]
        results = list(executor.map(batch_run_agent_replica, agent_replicas, replica_tapes))
        final_tapes = list(chain(*[r[1] for r in results]))
        agent_replicas = [r[0] for r in results]  # type: ignore
    logger.info(f"Making tapes took {time.time() - start_making_tapes}")

    for new_tape in tqdm(final_tapes, total=len(final_tapes), desc="Extracting training data from tapes", unit="tape"):
        tape_training_samples, tape_stats = extract_tape_training_samples(new_tape, agent_replicas[0], cfg)
        training_samples.extend(tape_training_samples)
        reward_stats[new_tape.metadata.parent_id].append(tape_stats["reward"])
        success_stats[new_tape.metadata.parent_id].append(tape_stats["success"])
        no_errors_stats[new_tape.metadata.parent_id].append(tape_stats["no_error"])
        prompt_tokens_stats[new_tape.metadata.parent_id].append(tape_stats["prompt_tokens"])
        output_tokens_stats[new_tape.metadata.parent_id].append(tape_stats["output_tokens"])
        overflow_stats[new_tape.metadata.parent_id].append(tape_stats["overflows"])

    end_make_data = time.time()

    stats = {
        **{f"{split_name}_{k}_reward": v for k, v in calculate_stats(reward_stats).items()},
        **{f"{split_name}_{k}_success": v for k, v in calculate_stats(success_stats).items()},
        **{f"{split_name}_{k}_no_errors": v for k, v in calculate_stats(no_errors_stats).items()},
        **{f"{split_name}_{k}_prompt_tokens": v for k, v in calculate_stats(prompt_tokens_stats).items()},
        **{f"{split_name}_{k}_output_tokens": v for k, v in calculate_stats(output_tokens_stats).items()},
        **{
            f"execution_time/{split_name}_make_data": end_make_data - start_make_data,
            f"execution_time/{split_name}_tapes_made_per_second": len(final_tapes) / (end_make_data - start_make_data),
            f"{split_name}_prompt_tokens": sum([sum(pt) for pt in prompt_tokens_stats.values()]),
            f"{split_name}_output_tokens": sum([sum(ot) for ot in output_tokens_stats.values()]),
            f"{split_name}_overflows": np.mean([np.mean(ov) for ov in overflow_stats.values()]),
        },
    }
    return agent_replicas, final_tapes, training_samples, stats


@hydra.main(config_path="../../conf/", config_name="rl_gsm8k", version_base="1.3.2")
def main(cfg: DictConfig):
    if cfg.llm.parameters.temperature != 1.0:
        raise ValueError("Temperature must be 1.0 for RL training")

    multiprocessing.set_start_method("spawn")  # necessary to use gpus in subprocesses
    random.seed(42)
    exp_path = Path(cfg.output_dir)
    setup_logging(exp_path)
    logger.info(f"Current dir: {os.getcwd()}, output dir: {cfg.output_dir}")
    cfg.finetune.wandb_id = str(exp_path).replace("/", "_")
    run = init_wandb(cfg, exp_path, flatten_dict_config(cfg))
    if run is None:
        raise ValueError("Failed to initialize wandb run")
    state_path = exp_path / "rl_state.json"
    state = load_state(state_path)
    # optionally clean all data at start time
    if cfg.force_restart:
        clean_up(exp_path, state, state_path)

    train_samples, test_samples = load_datasets()
    conf_dir = exp_path / "conf"
    os.makedirs(conf_dir, exist_ok=True)
    finetune_path = exp_path / "finetune"

    while state["iteration"] < cfg.max_iterations:
        logger.info(f"Starting iteration {state['iteration']}")
        start_iteration = time.time()
        if os.path.exists(finetune_path / "current"):
            assistant_model_path = str(finetune_path / "current")
        else:
            assistant_model_path = cfg.model_path

        try:
            all_results = {}
            # Collect tapes using the assistant model
            # We might also evaluate the assistant model on the test set
            with VLLMServiceManager(
                exp_path=exp_path,
                service_name="actor",
                model_name_or_path=assistant_model_path,
                port=8080,
                verbose=True,
                cuda_device=",".join([str(i) for i in range(torch.cuda.device_count())]),
                **(dict(cfg.vllm_config.vllm_kwargs) | dict(cfg.vllm_config.actor_vllm_kwargs)),
            ) as vllm_service_manager:
                sub_samples = random.sample(train_samples, cfg.max_agent_forks // cfg.attempts)
                train_tapes = convert_problems_to_tapes(sub_samples, cfg)
                train_tapes = [copy.deepcopy(tape) for tape in train_tapes for _ in range(cfg.attempts)]
                train_llms = [
                    TrainableLLM(
                        base_url=base_url,
                        model_name=str(assistant_model_path),
                        tokenizer_name=str(assistant_model_path),
                        parameters=cfg.llm.parameters,
                        use_cache=False,
                        collect_logprobs=True,
                        observe_llm_calls=False,
                    )
                    for base_url in vllm_service_manager.get_base_urls()
                ]

                # one agent per LLM
                train_agent_replicas = [
                    CoTMathAgent.create(
                        system_prompt=cfg.system_prompt, llm=llm, max_prompt_length=cfg.max_prompt_length
                    )
                    for llm in train_llms
                ]

                splits = [("train", train_agent_replicas, train_tapes)]
                if state["iteration"] % cfg.test_every_n_iterations == 0 and cfg.test_every_n_iterations > 0:
                    # Create test agent replicas with different LLM parameters
                    test_tapes = convert_problems_to_tapes(test_samples, cfg)
                    test_llms = [
                        TrainableLLM(
                            base_url=base_url,
                            model_name=str(assistant_model_path),
                            tokenizer_name=str(assistant_model_path),
                            parameters=cfg.test_llm.parameters,
                            use_cache=False,
                            collect_logprobs=True,
                            observe_llm_calls=False,
                        )
                        for base_url in vllm_service_manager.get_base_urls()
                    ]
                    test_agent_replicas = [
                        CoTMathAgent.create(
                            system_prompt=cfg.system_prompt, llm=llm, max_prompt_length=cfg.max_prompt_length
                        )
                        for llm in test_llms
                    ]
                    splits.append(("test", test_agent_replicas, test_tapes))

                for split_name, agent_replicas, tapes in splits:
                    tapes_dir = exp_path / "tapes" / split_name / str(state["iteration"])
                    agent_replicas_with_stats, new_tapes, split_training_samples, stats = generate_training_data(
                        agent_replicas, tapes, cfg, tapes_dir, split_name
                    )

                    llm_stats = agent_replicas_with_stats[0].llm.get_stats()
                    make_data_took = stats[f"execution_time/{split_name}_make_data"]
                    llm_stats = {f"llm/{split_name}_{k}": v for k, v in llm_stats.items()}
                    throughput_stats = {
                        f"{split_name}_prompt_tokens_per_sec": stats[f"{split_name}_prompt_tokens"] / make_data_took,
                        f"{split_name}_output_tokens_per_sec": stats[f"{split_name}_output_tokens"] / make_data_took,
                        f"{split_name}_total_tokens_per_sec": (
                            stats[f"{split_name}_prompt_tokens"] + stats[f"{split_name}_output_tokens"]
                        )
                        / make_data_took,
                    }
                    stats.update(llm_stats)
                    stats.update(throughput_stats)

                    all_results[split_name] = {
                        "new_tapes": new_tapes,
                        "training_samples": split_training_samples,
                        "stats": stats,
                    }

                    # Log results
                    logger.info(f"{split_name} stats:")
                    for stat_name, stat_value in stats.items():
                        logger.info(f"{stat_name}: {stat_value}")
                assistant_vllm_stats = vllm_service_manager.get_stats()

        except Exception as e:
            logger.error(colored(f"Failed to solve task: {e}", "red"))
            raise e

        training_samples = all_results["train"]["training_samples"]
        logger.info(f"Collected {len(training_samples)} training samples")
        stats = all_results["train"]["stats"]
        if "test" in all_results:  # test is only present every cfg.test_every_n_iterations
            stats.update(all_results["test"]["stats"])
            time_evaluation = stats["execution_time/test_make_data"]
        else:
            time_evaluation = 0
        wandb.log(
            stats,
            step=state["iteration"],
        )

        try:
            # Populate reference logprobs using the reference model
            with VLLMServiceManager(
                exp_path=exp_path,
                service_name="reference",
                model_name_or_path=cfg.model_path,
                port=8180,
                verbose=True,
                cuda_device=",".join([str(i) for i in range(torch.cuda.device_count())]),
                **(dict(cfg.vllm_config.vllm_kwargs) | dict(cfg.vllm_config.ref_vllm_kwargs)),
            ) as vllm_service_manager:
                ref_llms = [
                    TrainableLLM(
                        base_url=url,
                        model_name=cfg.model_path,
                        tokenizer_name=cfg.model_path,
                        parameters=dict(temperature=0.7),
                    )
                    for url in vllm_service_manager.get_base_urls()
                ]

                start_basemodel_logprobs = time.time()
                with ThreadPoolExecutor(
                    max_workers=cfg.get_logprobs_workers_per_gpu * torch.cuda.device_count()
                ) as executor:
                    chunk_size = 64
                    futures = []
                    for chunk_id, chunk_offset in enumerate(range(0, len(training_samples), chunk_size)):
                        ref_llm = ref_llms[chunk_id % len(ref_llms)]
                        chunk = training_samples[chunk_offset : chunk_offset + chunk_size]
                        futures.append(executor.submit(batch_annotate_traces_with_ref_logprobs, ref_llm, chunk))
                    # Reference logprobs are added in-place
                    futures = tqdm(as_completed(futures), total=len(futures), desc="Adding logprobs")
                    _ = [future.result() for future in futures]

                refmodel_vllm_stats = vllm_service_manager.get_stats()
                refmodel_starting_time = refmodel_vllm_stats["starting_time"]
                time_populating_ref_logprobs = time.time() - start_basemodel_logprobs

        except Exception as e:
            logger.error(colored(f"Failed to get ref log probs: {e}", "red"))
            raise e

        logprob_stats = {
            "execution_time/populating_ref_logprobs": time_populating_ref_logprobs,
            "execution_time/starting_assistantmodel_vllm": assistant_vllm_stats["starting_time"],
            "execution_time/starting_refmodel_vllm": refmodel_starting_time,
        }
        logger.info("Logprob population stats:")
        for stat_name, stat_value in logprob_stats.items():
            logger.info(f"{stat_name}: {stat_value}")
        wandb.log(logprob_stats, step=state["iteration"])
        rollout_dir = exp_path / "rollouts" / str(state["iteration"])
        os.makedirs(rollout_dir, exist_ok=True)
        with open(rollout_dir / "data.jsonl", "w") as f:
            for trace in training_samples:
                f.write(trace.model_dump_json() + "\n")
                f.flush()

        # Create a config for this finetuning iteration
        finetune_cfg = cfg.copy()

        # we increment the number of steps to interrupt the training
        checkpoint_steps = finetune_cfg.finetune.save_checkpoint_steps
        interrupt_train_steps = int((state["iteration"] + 1) * checkpoint_steps)

        finetune_cfg.finetune.interrupt_train_steps = interrupt_train_steps
        finetune_cfg.output_dir = str(finetune_path)
        finetune_cfg.finetune.data = {"data_parts_train": [{"path": str(rollout_dir)}]}
        finetune_cfg.finetune.wandb_id = run.id + "_finetune"
        finetune_cfg.finetune.wandb_name = run.name + "_finetune"
        finetune_cfg.finetune.wandb_resume = "always"
        config_path = conf_dir / f"{state['iteration']}.yaml"
        OmegaConf.save(finetune_cfg, config_path)

        start_finetune = time.time()
        # launch the finetuning in a subprocess
        launch_training(
            str(conf_dir),
            str(state["iteration"]),
            cfg.accelerate_cfg_path,
            use_deepspeed=cfg.use_deepspeed,  # defaults to False
        )
        time_finetune = time.time() - start_finetune
        time_iteration = time.time() - start_iteration
        wandb.log(
            {
                "execution_time/finetune": time_finetune,
                "execution_time/iteration": time_iteration,
                "execution_time/overhead": time_iteration
                - time_finetune
                - time_populating_ref_logprobs
                - time_evaluation
                - stats["execution_time/train_make_data"],
            },
            step=state["iteration"],
        )
        state["iteration"] += 1
        save_state(state, state_path)

    logger.info(f'Finished training after {state["iteration"]} iterations')


if __name__ == "__main__":
    main()



"""Batch processing of tapes."""

import logging
import random
import traceback
from functools import partial
from pathlib import Path
from typing import Generator, Generic, Sequence

from pydantic import BaseModel

from tapeagents.agent import Agent, Annotator, ObservationMaker
from tapeagents.config import is_debug_mode
from tapeagents.core import AnnotatorTapeType, ObservationMakerTapeType, Tape, TapeMetadata, TapeType
from tapeagents.environment import Environment
from tapeagents.io import stream_yaml_tapes
from tapeagents.orchestrator import main_loop
from tapeagents.parallel_processing import choose_processor

logger = logging.getLogger(__name__)

_DEFAULT_N_WORKERS = 16


def batch_main_loop(
    agent: Agent[TapeType],
    tapes: list[TapeType],
    environments: Environment | list[Environment],
    n_workers: int = _DEFAULT_N_WORKERS,
    strict: bool = False,
    max_loops: int = -1,
) -> Generator[TapeType, None, None]:
    """Continue tapes in parallel using an agent."""
    if not isinstance(environments, list):
        environments = [environments] * len(tapes)

    def worker_func(
        input: tuple[TapeType, Environment], agent: Agent, max_loops: int, strict: bool
    ) -> TapeType | Exception:
        start_tape, env = input
        try:
            result = main_loop(agent, start_tape, env, max_loops=max_loops).get_final_tape()
        except Exception as e:
            if is_debug_mode() or strict:
                return e
            return start_tape.model_copy(
                update=dict(metadata=TapeMetadata(parent_id=start_tape.metadata.id, error=traceback.format_exc()))
            )
        result.metadata.parent_id = start_tape.metadata.id
        return result

    processor = choose_processor(n_workers=n_workers)
    worker_func_partial = partial(worker_func, agent=agent, max_loops=max_loops, strict=strict)
    for smth in processor(zip(tapes, environments), worker_func_partial):
        if isinstance(smth, Tape):
            yield smth
        else:
            raise smth


class ObsLayerConfig(BaseModel, Generic[TapeType, ObservationMakerTapeType]):
    obs_makers: list[tuple[ObservationMaker[TapeType, ObservationMakerTapeType], int]]


def batch_add_observations(
    tapes: list[TapeType],
    layer_conf: ObsLayerConfig[TapeType, ObservationMakerTapeType],
    n_workers: int = _DEFAULT_N_WORKERS,
    seed: int = 1,
    strict: bool = False,
    shuffle: bool = True,
) -> Generator[tuple[TapeType, ObservationMakerTapeType], None, None]:
    # TODO: add environments, use main loop
    def worker_func(input_: tuple[TapeType, ObservationMaker]) -> tuple[TapeType, ObservationMakerTapeType] | Exception:
        try:
            tape, obs_maker = input_
            for event in obs_maker.run(obs_maker.make_own_tape(tape)):
                if event.final_tape:
                    return obs_maker.add_observation(tape, event.final_tape), event.final_tape
        except Exception as e:
            logger.error(e, exc_info=True)
            if is_debug_mode() or strict:
                return e
            result_tape = tape.model_copy(update=dict(metadata=TapeMetadata(parent_id=tape.metadata.id, error=repr(e))))
            result_obs_maker_tape = obs_maker.make_own_tape(tape)
            result_obs_maker_tape.metadata.error = str(e)
            return result_tape, result_obs_maker_tape
        raise Exception("Observation maker did not finish the tape")

    # TODO: generate for many obs_makers at the same time
    rng = random.Random(seed)
    for i, (obs_maker, count) in enumerate(layer_conf.obs_makers):
        possible_tapes = [t for t in tapes if obs_maker.can_continue(t)]
        # TODO: what if len(possible_tapes) == 0?
        if shuffle:
            # take 'count' random tapes from possible_tapes using seed
            chosen_tapes = rng.choices(possible_tapes, k=count)
        else:
            # TODO: not fully deterministic because of obs_maker.can_continue()
            chosen_tapes = [possible_tapes[i % len(possible_tapes)] for i in range(count)]

        processor = choose_processor(n_workers=n_workers)
        for smth in processor(zip(chosen_tapes, [obs_maker] * count), worker_func):
            if isinstance(smth, tuple):
                yield smth
            else:
                raise smth


def batch_annotate(
    tapes: list[TapeType],
    annotator: Annotator[TapeType, AnnotatorTapeType],
    n_workers: int = _DEFAULT_N_WORKERS,
    strict: bool = False,
) -> Generator[AnnotatorTapeType, None, None]:
    def worker_func(input_: tuple[TapeType, Annotator]) -> AnnotatorTapeType | Exception:
        try:
            tape, annotator = input_
            return annotator.annotate(tape)
        except Exception as e:
            if is_debug_mode() or strict:
                return e
            result = annotator.make_own_tape(tape)
            result.metadata.error = str(e)
            return result

    processor = choose_processor(n_workers=n_workers)
    for smth in processor(zip(tapes, [annotator] * len(tapes)), worker_func):
        if isinstance(smth, Tape):
            yield smth
        else:
            raise smth


def generate_tapes(
    agent: Agent[TapeType],
    start_tapes: list[TapeType],
    environments: Environment | list[Environment],
    layers_conf: Sequence[ObsLayerConfig[TapeType, ObservationMakerTapeType]],
    work_dir: Path,
    annotator: Annotator[TapeType, AnnotatorTapeType] | None = None,
    n_workers: int = _DEFAULT_N_WORKERS,
    seed: int = 1,
    strict: bool = False,
    shuffle: bool = True,
) -> list[list[TapeType]]:
    tapes = start_tapes
    layer_tapes = [[]] * (len(layers_conf) * 2)

    for i, layer_conf in enumerate(layers_conf):
        begin_tapes_path = work_dir / f"tapes_{i}_begin.yaml"
        end_tapes_path = work_dir / f"tapes_{i}_end.yaml"
        obs_maker_tapes_path = work_dir / f"obs_maker_tapes_{i}.yaml"
        annotator_tapes_path = work_dir / f"annotator_tapes_{i}.yaml"
        logger.info(f"tapes: {len(tapes)}")

        # make observations
        new_layer_begin = []
        with (
            stream_yaml_tapes(begin_tapes_path) as tapes_dumper,
            stream_yaml_tapes(obs_maker_tapes_path) as obs_maker_tapes_dumper,
        ):
            for tape, obs_maker_tape in batch_add_observations(
                tapes, layer_conf, n_workers=n_workers, seed=seed, strict=strict, shuffle=shuffle
            ):
                tapes_dumper.save(tape)
                obs_maker_tapes_dumper.save(obs_maker_tape)
                if tape.metadata.error is None:
                    new_layer_begin.append(tape)
        logger.info(f"new_layer_begin: {len(new_layer_begin)} tapes")

        # run the agent
        new_layer_end = []
        with stream_yaml_tapes(end_tapes_path) as dumper:
            for tape in batch_main_loop(agent, new_layer_begin, environments, n_workers=n_workers, strict=strict):
                dumper.save(tape)
                if tape.metadata.error is None:
                    new_layer_end.append(tape)

        # run the annotator
        with stream_yaml_tapes(annotator_tapes_path) as dumper:
            if annotator is not None:
                for tape in batch_annotate(new_layer_end, annotator, n_workers=n_workers, strict=strict):
                    dumper.save(tape)

        logger.info(f"new_layer_end: {len(new_layer_end)} tapes")
        tapes = new_layer_end
        layer_tapes[2 * i] = new_layer_begin
        layer_tapes[2 * i + 1] = new_layer_end

    return layer_tapes



import asyncio
import time
from unittest.mock import MagicMock

import aiohttp
import pytest

from tapeagents.agent import Agent
from tapeagents.core import Prompt
from tapeagents.dialog_tape import DialogTape, UserStep
from tapeagents.llms import LLMEvent, LLMOutput, LLMStream, TrainableLLM
from tapeagents.nodes import StandardNode
from tapeagents.steps import ReasoningThought


class MockAsyncLLM(TrainableLLM):
    """Mock Async LLM interface with 1 second delay in responses"""

    async def agenerate(self, prompt: Prompt, session: aiohttp.ClientSession, **kwargs) -> LLMStream:
        """Simulate LLM response with 1 second delay"""
        await asyncio.sleep(1)  # Simulate 1 second delay

        # Return different responses based on the question
        if "capital of France" in prompt.messages[0]["content"]:
            content = "The capital of France is Paris."
        elif "capital of Germany" in prompt.messages[0]["content"]:
            content = "The capital of Germany is Berlin."
        else:
            content = f"Response to: {prompt}"
        output = LLMOutput(content=content)
        llm_call = self.log_output(prompt, output)

        def _gen():
            yield LLMEvent(llm_call=llm_call, output=output)

        return LLMStream(generator=_gen(), prompt=prompt)

    def count_tokens(self, messages: list[dict] | str) -> int:
        return 1


@pytest.mark.asyncio
async def test_agent_arun():
    """Test the arun method of Agent class"""
    # Create agent with mocked LLM
    llm = MockAsyncLLM(model_name="mock-model")
    agent = Agent.create(
        llms=llm,
        nodes=[
            StandardNode(guidance="lets think step by step"),
            StandardNode(name="verify", guidance="lets verify the answer"),
        ],
    )

    # Create test question and tape
    question = "What is the capital of France?"
    tape = DialogTape(steps=[UserStep(content=question)])

    # Test arun method
    events = []
    mock_session = MagicMock()
    async for event in agent.arun(tape=tape, session=mock_session, max_iterations=2):
        events.append(event)

    # Verify results
    assert len(events) > 2
    step_events = [e for e in events if e.step]
    assert len(step_events) == 2

    # Check both nodes were used
    node_names = [e.step.metadata.node for e in step_events]
    assert "StandardNode" in node_names
    assert "verify" in node_names

    # Check response content
    assert isinstance(step_events[0].step, ReasoningThought)
    assert isinstance(step_events[1].step, ReasoningThought)
    assert "Paris" in step_events[0].step.reasoning
    assert "Paris" in step_events[1].step.reasoning


@pytest.mark.asyncio
async def test_agent_arun_iteration():
    """Test the arun_iteration method of Agent class"""
    # Create agent with mocked LLM
    llm = MockAsyncLLM(model_name="mock-model")
    agent = Agent.create(
        llms=llm,
        nodes=[
            StandardNode(guidance="lets think step by step"),
            StandardNode(name="verify", guidance="lets verify the answer"),
        ],
    )

    # Create test question and tape
    question = "What is the capital of France?"
    tape = DialogTape(steps=[UserStep(content=question)])

    # Test single iteration
    mock_session = MagicMock()
    steps = []
    async for step in agent.arun_iteration(tape=tape, session=mock_session):
        steps.append(step)

    assert len(steps) == 1

    # Verify results
    step = steps[0]
    assert step is not None
    assert step.metadata.node == "StandardNode"
    assert isinstance(step, ReasoningThought)
    assert "Paris" in step.reasoning


@pytest.mark.asyncio
async def test_concurrent_execution():
    """Test concurrent execution of multiple questions"""
    # Create agent with mocked LLM
    llm = MockAsyncLLM(model_name="mock-model")
    agent = Agent.create(
        llms=llm,
        nodes=[
            StandardNode(guidance="lets think step by step"),
            StandardNode(name="verify", guidance="lets verify the answer"),
        ],
    )

    # Set up multiple questions
    questions = [
        "What is the capital of France?",
        "What is the capital of Germany?",
        "How are you?",
        "What is the capital of Italy?" "What is the capital of Spain?",
        "Who is the president of the USA?",
        "How is the weather today?",
        "Can you tell me a joke?",
        "What is the meaning of life?",
        "Hey, what is your name?",
        "What is the capital of Japan?",
    ]
    mock_session = MagicMock()

    # Function to process one question
    async def process_question(question):
        steps = []
        async for event in agent.arun(
            tape=DialogTape(steps=[UserStep(content=question)]),
            session=mock_session,
            max_iterations=2,
        ):
            if event.step:
                steps.append(event.step)
        return question, steps

    # Measure time for concurrent execution
    start_time = time.time()
    results = await asyncio.gather(*[process_question(q) for q in questions])
    elapsed_time = time.time() - start_time

    # Verify results
    assert len(results) == 10

    # Check if both questions were processed correctly
    for question, steps in results:
        assert len(steps) == 2
        if "France" in question:
            assert "Paris" in steps[0].reasoning
            assert "Paris" in steps[1].reasoning
        elif "Germany" in question:
            assert "Berlin" in steps[0].reasoning
            assert "Berlin" in steps[1].reasoning

    assert elapsed_time < 3  # 2 llm requests per question each taking 1 second, plus some overhead
    print(f"Elapsed time {elapsed_time:.2f}")


if __name__ == "__main__":
    asyncio.run(test_concurrent_execution())



"""
GUI for browsing tapes.
"""

import datetime
import logging
import os

import gradio as gr
import uvicorn
import yaml
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from tapeagents.core import Tape, TapeMetadata
from tapeagents.io import load_tapes, stream_yaml_tapes
from tapeagents.observe import retrieve_tape_llm_calls
from tapeagents.renderers.basic import BasicRenderer
from tapeagents.steps import Annotation

logger = logging.getLogger(__name__)


class TapeBrowser:
    """
    Base class for tape browser GUI.
    Displays the list of tapes from the files in the folder, allows to select the file with a bunch of tapes
    and then select tape and view it. Allows to navigate between parent and children tapes by links.
    """

    def __init__(
        self,
        tape_cls: type[Tape],
        tapes_folder: str,
        renderer: BasicRenderer,
        file_extension: str = ".yaml",
    ):
        self.tape_cls = tape_cls
        self.file_extension: str = file_extension
        self.tapes_folder: str = tapes_folder
        self.annotation_file: str = os.path.join(tapes_folder, "annotations.yaml")
        self.renderer: BasicRenderer = renderer
        self.files: list[str] = []
        self.tape_index: dict[str, tuple[int, int]] = {}
        self.tape_children: dict[str, list[str]] = {}
        self.request: gr.Request | None = None
        self.selected_tape: int = 0
        self.tapes: list[Tape] = []
        self.llm_calls: dict = {}

    def load_tapes(self, fname: str) -> list[Tape]:
        fpath = os.path.join(self.tapes_folder, fname)
        try:
            tapes = load_tapes(self.tape_cls, fpath)
            logger.info(f"{len(tapes)} tapes loaded from {fname}")
        except Exception:
            logger.error(f"Could not load tapes from {fpath}")
            raise
        return tapes

    def load_llm_calls(self):
        logger.info("Loading LLM calls")
        self.llm_calls = retrieve_tape_llm_calls(self.tapes)
        logger.info(f"Loaded {len(self.llm_calls)} LLM calls")

    def get_steps(self, tape: Tape) -> list:
        return tape.steps

    def get_context(self, tape: Tape) -> list:
        return getattr(tape.context, "steps", [])

    def get_tape_files(self) -> list[str]:
        files = sorted([f for f in os.listdir(self.tapes_folder) if f.endswith(self.file_extension)])
        assert files, f"No files found in {self.tapes_folder}"
        logger.info(f"{len(files)} files found in {self.tapes_folder}")
        indexed = 0
        nonempty_files = []
        for i, file in enumerate(files):
            if file in self.files:
                continue  # already indexed
            tapes = self.load_tapes(file)
            if not len(tapes):
                logger.warning(f"File {file} does not contain any known tapes, skip")
                continue
            for j, tape in enumerate(tapes):
                tape_id = tape.metadata.id
                parent_id = tape.metadata.parent_id
                if tape_id:
                    if tape_id in self.tape_index and self.tape_index[tape_id] != (i, j):
                        raise ValueError(
                            f"Duplicate tape id {tape_id}. Both in {self.tape_index[tape_id]} and {(i, j)}"
                        )
                    indexed += 1
                    self.tape_index[tape_id] = (i, j)
                    if parent_id:
                        if parent_id not in self.tape_children:
                            self.tape_children[parent_id] = []
                        self.tape_children[parent_id].append(tape_id)
            nonempty_files.append(file)
        logger.info(f"Indexed {indexed} new tapes, index size: {len(self.tape_index)}")
        logger.info(f"{len(self.tape_children)} tapes with children found")
        return nonempty_files

    def get_tape_label(self, tape: Tape) -> str:
        label = ""
        parent_tape_id = tape.metadata.parent_id
        tape_id = tape.metadata.id
        if tape_id:
            label += f"<br><b>ID: {tape_id}</b>"
        if parent_tape_id:
            if parent_tape_id in self.tape_index:
                logger.info(f"Tape index value {self.tape_index[parent_tape_id]}")
                fid, tid = self.tape_index[parent_tape_id]
                logger.info(f"Parent tape {parent_tape_id} found in {self.files[fid]}")
                tape_name = f"{self.files[fid]}/{tid}"
                label += f'<br>Parent Tape: <a href="?tape_id={parent_tape_id}">{tape_name}</a>'
            else:
                label += f"<br>Parent tape: {parent_tape_id}"
        if tape_id and tape_id in self.tape_children:
            children = []
            for cid in self.tape_children[tape_id]:
                tape_name = f"{self.files[self.tape_index[cid][0]]}/{self.tape_index[cid][1]}"
                children.append(f'<div style="margin-left:2em;"><a href="?tape_id={cid}">{tape_name}</a></div>')
            label += f"<br>Children tapes:<br>{''.join(children)}"
        label += f"<br>Length: {len(tape)} steps"
        if tape.metadata:
            m = {k: v for k, v in tape.metadata.model_dump().items() if k not in ["id", "parent_id"]}
            label += f'<h3>Metadata</h3><div style="white-space: pre-wrap;">{yaml.dump(m, allow_unicode=True)}</div>'
        return label

    def get_exp_label(self, filename: str, tapes: list[Tape]) -> str:
        tapelengths = [len(tape) for tape in tapes]
        tapelen = sum(tapelengths) / len(tapelengths)
        return f"<h3>{len(self.tape_index)} indexed tapes in {len(self.files)} files<br>{len(tapes)} tapes in the current file<br>Avg. tape length: {tapelen:.1f} steps</h3>"

    def get_tape_name(self, i: int, tape: Tape) -> str:
        return f"Tape {i}"

    def update_view(self, selected_file: str):
        logger.info(f"Loading tapes from {selected_file}")
        self.tapes = self.load_tapes(selected_file)
        self.load_llm_calls()
        file_label = self.get_exp_label(selected_file, self.tapes)
        tape_names = [(self.get_tape_name(i, tape), i) for i, tape in enumerate(self.tapes)]
        logger.info(f"Selected file: {selected_file}, selected tape: {self.selected_tape}")
        files = gr.Dropdown(self.files, label="Experiment", value=selected_file)  # type: ignore
        tape_names = gr.Dropdown(tape_names, label="Tape", value=self.selected_tape)  # type: ignore
        tape_html, label = self.update_tape_view(self.selected_tape)
        return files, tape_names, file_label, tape_html, label

    def update_tape_view(self, tape_id: int) -> tuple[str, str]:
        logger.info(f"Loading tape {tape_id}")
        if tape_id >= len(self.tapes):
            logger.error(f"Tape {tape_id} not found in the index")
            return f"<h1>Failed to load tape {tape_id}</h1>", ""
        tape = self.tapes[tape_id]
        label = self.get_tape_label(tape)
        html = f"{self.renderer.style}{self.renderer.render_tape(tape, self.llm_calls)}"
        return html, label

    def reload_tapes(self, selected_file: str):
        logger.info(f"Reloading tapes from {selected_file}")
        return self.update_view(selected_file)

    def switch_file(self, selected_file: str):
        logger.info(f"Switching to file {selected_file}")
        self.selected_tape = 0
        return self.update_view(selected_file)

    def update_view_from_request(self, request: gr.Request):
        self.request = request
        tape_id = self.request.query_params.get("tape_id")
        self.files = self.get_tape_files()
        selected_file = self.files[0]
        if tape_id and tape_id in self.tape_index:
            i, j = self.tape_index[tape_id]
            selected_file = self.files[i]
            self.selected_tape = j
            logger.info(f"Selected tape {selected_file}/{j} from query params")
        return self.update_view(selected_file)

    def save_annotation(self, step: int, annotation: str, tape_id: int):
        tape = self.tapes[tape_id]
        time_now = datetime.datetime.now().isoformat()
        annotator_tape = Tape[Tape, Annotation](
            metadata=TapeMetadata(other={"made_at": time_now}),
            context=tape,
            steps=[Annotation(step=step, text=annotation)],
        )
        with stream_yaml_tapes(self.annotation_file, "a") as saver:
            saver.save(annotator_tape)
        gr.Info(f'Saved an annotation to "{self.annotation_file}')

    def create_blocks(self):
        with gr.Blocks(analytics_enabled=False, title="TapeAgents Browser") as blocks:
            with gr.Row():
                with gr.Column(scale=1):
                    file_selector = gr.Dropdown([], label="File")
                    reload_button = gr.Button("Reload Tapes")
                    file_label = gr.HTML("")
                with gr.Column(scale=4):
                    tape_view = gr.HTML("")
                with gr.Column(scale=1):
                    tape_selector = gr.Dropdown([], label="Tape")
                    tape_label = gr.HTML("")
                    gr.HTML("<h2> Annotate tape </h2>")
                    step_selector = gr.Number(0, label="Step")
                    annotation_text = gr.Textbox("", label="Annotation Text", lines=5)
                    annotate_button = gr.Button("Annotate")
            reload_button.click(
                fn=self.reload_tapes,
                inputs=[file_selector],
                outputs=[file_selector, tape_selector, file_label, tape_view, tape_label],
            )
            tape_selector.input(fn=self.update_tape_view, inputs=tape_selector, outputs=[tape_view, tape_label])
            file_selector.input(
                fn=self.switch_file,
                inputs=file_selector,
                outputs=[file_selector, tape_selector, file_label, tape_view, tape_label],
            )
            annotate_button.click(
                fn=self.save_annotation, inputs=[step_selector, annotation_text, tape_selector], outputs=[]
            )
            blocks.load(
                self.update_view_from_request,
                None,
                outputs=[file_selector, tape_selector, file_label, tape_view, tape_label],
            )
        return blocks

    def launch(self, server_name: str = "0.0.0.0", port=7860, debug: bool = False, static_dir: str = ""):
        self.blocks = self.create_blocks()

        if static_dir:
            logger.info(f"Starting FastAPI server with static dir {static_dir}")
            # mount Gradio app to FastAPI app
            app = FastAPI()
            app.mount("/static", StaticFiles(directory=static_dir), name="static")
            app = gr.mount_gradio_app(app, self.blocks, path="/")
            uvicorn.run(app, host=server_name, port=port)
        else:
            self.blocks.launch(server_name=server_name, server_port=port, debug=debug)



import json
import logging
import os
import sys

from tqdm import tqdm

from tapeagents.core import TrainingText
from tapeagents.finetune.data import load_samples, save_samples
from tapeagents.finetune.finetune import load_config, run_finetuning_loop
from tapeagents.llms import TrainableLLM

from .math_agent import ActionExecutionFailure, MathAgent, MathTape

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_training_samples_from_tapes(tapes_path: str) -> list[TrainingText]:
    """
    Make training samples from tapes that were solved successfully,
    does not contain action execution failures, and do not have repeated steps.
    """

    # We need the agent to cut tapes into training samples
    agent = MathAgent.create(
        TrainableLLM(
            base_url="https://api.together.xyz",
            model_name="meta-llama/Meta-Llama-3.1-8B-Instruct-Turbo",
            tokenizer_name="meta-llama/Meta-Llama-3.1-8B-Instruct",
        )
    )

    training_samples: list[TrainingText] = []
    failures = 0
    not_solved = 0
    with_duplicates = 0
    tape_files = sorted([os.path.join(tapes_path, f) for f in os.listdir(tapes_path) if f.endswith(".json")])
    for tape_file in tqdm(tape_files):
        with open(tape_file) as f:
            tape_dict = json.load(f)
        tape = MathTape.model_validate(tape_dict)
        step_types = set(type(step) for step in tape)

        # detect repeated steps
        last_step_view = None
        duplicate = False
        for step in tape:
            view = step.llm_view()
            if view == last_step_view:
                duplicate = True
                break
            last_step_view = view

        if duplicate:
            with_duplicates += 1
            continue
        if ActionExecutionFailure in step_types:
            failures += 1
            continue
        if not tape.metadata.result["solved"]:
            not_solved += 1
            continue

        for sample in agent.make_training_data(tape):
            training_samples.append(sample)
    logger.info(f"Skipped failures {failures}, not solved {not_solved}, with duplicates {with_duplicates}")
    logger.info(f"Created positive training samples: {len(training_samples)}")
    return training_samples


def main(exp_path: str):
    train_samples_file = "gsm8k/tuning/llama31_70b_train_t02/training_samples_3k.jsonl"
    if os.path.exists(train_samples_file):
        training_samples = load_samples(train_samples_file)
    else:
        tapes_path = "gsm8k/tuning/llama31_70b_train_t02/tapes"
        training_samples = get_training_samples_from_tapes(tapes_path)
        save_samples(training_samples, train_samples_file)

    cfg = load_config("llama31_8b", output_dir=exp_path)
    run_finetuning_loop(cfg=cfg, training_samples=training_samples)


if __name__ == "__main__":
    exp_path = sys.argv[1] if len(sys.argv) > 1 else "gsm8k/tuning/llama31_70b_train_t02/tune_llama31_8b_1"
    main(exp_path)



import json
import logging
import os
from collections import defaultdict
from pathlib import Path

import hydra
import torch
from browsergym.miniwob import ALL_MINIWOB_TASKS
from omegaconf import DictConfig
from termcolor import colored

from examples.rl_webagent.utils import VLLMServiceManager
from tapeagents.core import Action
from tapeagents.io import save_json_tape
from tapeagents.llms import TrainableLLM
from tapeagents.orchestrator import main_loop

from ..agent import WebAgent
from ..environment import WebEnvironment

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)


@hydra.main(
    version_base=None,
    config_path="../../../conf",
    config_name="webagent_demo",
)
def main(cfg: DictConfig) -> None:
    os.environ["TAPEAGENTS_SQLITE_DB"] = os.path.join(cfg.exp_path, "tapedata.sqlite")
    os.environ["MINIWOB_URL"] = cfg.environment_variables.miniwob_url
    # os.environ["SNOW_INSTANCE_URL"] = cfg.environment_variables.snow_instance_url
    # os.environ["SNOW_INSTANCE_UNAME"] = cfg.environment_variables.snow_instance_uname
    # os.environ["SNOW_INSTANCE_PWD"] = cfg.environment_variables.snow_instance_pwd

    tapes_dir = os.path.join(cfg.exp_path, "tapes")
    os.makedirs(tapes_dir, exist_ok=True)

    if os.path.exists(os.path.join(cfg.exp_path, "task_successes.json")):
        with open(os.path.join(cfg.exp_path, "task_successes.json"), "r") as f:
            task_successes = json.load(f)
    else:
        task_successes = defaultdict(list)

    with VLLMServiceManager(
        exp_path=Path(cfg.exp_path),
        service_name="actor",
        model_name_or_path=cfg.model_path,
        port=8080,
        verbose=True,
        cuda_device=",".join([str(i) for i in range(torch.cuda.device_count())]),
        **(dict(cfg.vllm_config.vllm_kwargs) | dict(cfg.vllm_config.actor_vllm_kwargs)),
    ) as vllm_service_manager:
        llms = [
            TrainableLLM(
                base_url=base_url,
                model_name=cfg.model_path,
                tokenizer_name=cfg.model_path,
                parameters=cfg.llm.parameters,
                use_cache=False,
                observe_llm_calls=False,
            )
            for base_url in vllm_service_manager.get_base_urls()
        ]
        env = WebEnvironment(**cfg.env)
        agent = WebAgent.create(llms[0])

        last_action = None
        repeated_action_cnt = 0
        for seed in cfg.seeds:
            for i, task in enumerate(ALL_MINIWOB_TASKS):
                task_name = f"task{i}_seed{seed}_{task.get_task_id()}"
                fname = f"{task_name}.json"
                if os.path.exists(os.path.join(tapes_dir, fname)):
                    logger.info(f"Skipping task {i+1}, already solved")
                    continue
                tmp_fpath = os.path.join(tapes_dir, f"{fname}.tmp")
                tape, metadata = env.start_task({"task": task, "seed": seed})
                metadata["seed"] = seed
                metadata["number"] = i
                logger.info(colored(f"Start task {i+1} seed {seed}: {metadata['goal']}", "cyan"))
                loop = 0
                logger.info(colored(f"Loop {loop+1}", "cyan"))
                try:
                    for event in main_loop(agent, tape, env, max_loops=20):  # type: ignore
                        if event.agent_event and event.agent_event.step:
                            step = event.agent_event.step
                            # avoid repeating the same action more than 4 times
                            if isinstance(step, Action):
                                step_view = step.llm_view()
                                if step_view == last_action:
                                    repeated_action_cnt += 1
                                    if repeated_action_cnt > 4:
                                        logger.error("Repeated action detected more than 4 time, stop the task")
                                        break
                                else:
                                    repeated_action_cnt = 0
                                last_action = step_view
                            tape = tape.append(step)  # type: ignore
                        if event.observation:
                            tape = tape.append(event.observation)  # type: ignore
                            loop += 1
                            logger.info(colored(f"Loop {loop+1}", "cyan"))
                        tape.metadata.result = metadata
                        save_json_tape(tape, tmp_fpath)
                    success, result = env.validate_task(tape)
                    final_reward = 1.0 if success else 0.0
                    tape.metadata.result.update({"success": success, **result, "final_reward": final_reward})
                except Exception as e:
                    logger.exception(e, stack_info=True)
                    logger.error(f"Failed to run task: {e}")
                    tape.metadata.result.update({"success": False, "reward": 0.0, "final_reward": 0.0, "error": str(e)})

                env.finish_task()
                os.unlink(tmp_fpath)  # remove temporary file
                logger.info(colored(f"reward: {tape.metadata.result['reward']}", "cyan"))
                save_json_tape(tape, tapes_dir, fname)
                logger.info(f"Saved tape to {fname}")

                # save task_successes
                if task.get_task_id() not in task_successes:
                    task_successes[task.get_task_id()] = []
                task_successes[task.get_task_id()].append(tape.metadata.result["reward"])

                overall_r = sum([sum(task_successes[task_id]) for task_id in task_successes]) / sum(
                    [len(task_successes[task_id]) for task_id in task_successes]
                )
                logger.info(colored(f"average reward: {overall_r}", "cyan"))

                with open(os.path.join(cfg.exp_path, "task_successes.json"), "w") as f:
                    json.dump(task_successes, f)

                # save average reward
                with open(os.path.join(cfg.exp_path, "task_avg_reward.json"), "w") as f:
                    json.dump(
                        {task_id: sum(rewards) / len(rewards) for task_id, rewards in task_successes.items()},
                        f,
                    )
                logger.info(f"Saved stats to {cfg.exp_path}")


if __name__ == "__main__":
    main()
    # to do RL, check examples/rl_gsm8k/orchestrate_rl.py
    # then, instead of run_batch() I need to create my own version to support agent & environemnt batch interactions
    # for that I can take inspiration from examples/gaia_agent/scripts/evaluate.py and look at:
    # Parallel(n_jobs=n_workers, prefer="processes")(
    #     [delayed(task_worker)(cfg, level, task_num) for level, task_num in tasks]
    # )



SYSTEM_PROMPT = """You are an expert AI Agent, your goal is to help the user perform tasks using a web browser.
Your role is to understand user queries and respond in a helpful and accurate manner.
Keep your replies concise and direct. Prioritize clarity and avoid over-elaboration.
You will be provided with the content of the current page and a task from the user.
Do not express your emotions or opinions about the user question."""

ALLOWED_STEPS = """
You are allowed to produce ONLY steps with the following json schemas:
{allowed_steps}
Do not reproduce schema when producing the steps, use it as a reference.
"""

hints = """
HINTS:
- You can use the BIDs of the elements to interact with them.
- To select value in the dropdown or combobox, ALWAYS use select_action.
- To click on the checkbox or radio button, ALWAYS use BID of corresponding LabelText and not the BID of the element itself.
- Press enter key to submit the search query.
- Always produce only one step at a time.
- Step kind is always lowercase and underscore separated.
"""

START = """
Produce reasoning_thought step that describes the intended solution to the task. In the reasoning lines:
- review the instructions from the user and the content of the page.
- outline the main task to be accomplished and the steps to be taken to achieve it.
- produce definiton of done, that will be checked later to verify if the task was completed.
Produce only one step!
"""

REFLECT = f"""
Review the current state of the page and previous steps to find the best possible next action to accomplish the task.
Produce reflection_thought.
{hints}
"""

ACT = f"""
Produce the next action to be performed with the current page.
If the task is already solved, produce final_answer_action and stop.
You can interact with the page elements using their BIDs as arguments for actions.
Be very cautious. Avoid submitting anything before verifying the effect of your
actions. Take the time to explore the effect of safe actions using reasoning first. For example
you can fill a few elements of a form, but don't click submit before verifying
that everything was filled correctly.
{hints}
"""

### Baseline Prompts ###

BASELINE_SYSTEM_PROMPT = """\
You are an agent trying to solve a web task based on the content of the page and
user instructions. You can interact with the page and explore, and send messages to the user. Each time you
submit an action it will be sent to the browser and you will receive a new page."""

GOAL_INSTRUCTIONS = """
# Instructions
Review the current state of the page and all other information to find the best
possible next action to accomplish your goal. Your answer will be interpreted
and executed by a program, make sure to follow the formatting instructions.

## Goal:
{goal}
"""

HINTS = """\
Note:
* Some tasks may be game like and may require to interact with the mouse position
in x, y coordinates.
* Some text field might have auto completion. To see it, you have to type a few
characters and wait until next step.
* If you have to cut and paste, don't forget to select the text first.
* Coordinate inside an SVG are relative to it's top left corner.
* Make sure to use bid to identify elements when using commands.
* Interacting with combobox, dropdowns and auto-complete fields can be tricky,
sometimes you need to use select_option, while other times you need to use fill
or click and wait for the reaction of the page.
"""

BE_CAUTIOUS = """
Be very cautious. Avoid submitting anything before verifying the effect of your
actions. Take the time to explore the effect of safe actions first. For example
you can fill a few elements of a form, but don't click submit before verifying
that everything was filled correctly.
"""

ABSTRACT_EXAMPLE = """
# Abstract Example

Here is an abstract version of the answer with description of the content of
each step. Only one single action can be executed after the thoughts. You can only use one action at a time.
Make sure you follow this structure, but replace the step fields content with your
answer:
[
    {
        "kind": "reasoning_thought",
        "reasoning": "<Think step by step. If you need to make calculations such as coordinates, write them here. Describe the effect that your previous action had on the current content of the page.>"
    },
    {
        "kind": "<some_of_the_allowed_action>",
        "<field_name>": "<field_content>"
    }
]
"""
CONCRETE_EXAMPLE = """
# Concrete Example

Here is a concrete example of how to format your answer.
Make sure to follow the template with proper steps:
[
    {
        "kind": "reasoning_thought",
        "reasoning": "From previous action I tried to set the value of year to "2022",\
using select_option, but it doesn't appear to be in the form. It may be a\
dynamic dropdown, I will try using click with the bid "a324" and look at the\
response from the page."
    },
    {
        "kind": "click_action",
        "bid": "a324"
    }
]
"""
BASELINE_STEPS_PROMPT = """# Action space:
Note: This action set allows you to interact with your environment. The primary way of referring to
elements in the page is through bid which are specified in your observations.
"""

MAC_HINT = "\nNote: you are on mac so you should use Meta instead of Control for Control+C etc.\n"



# Modified from the original source: https://github.com/microsoft/autogen/blob/gaia_multiagent_v01_march_1st/autogen/browser_utils.py
# MIT License

# Copyright (c) Microsoft Corporation.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.


import json
import logging
import mimetypes
import os
import pathlib
import re
import threading
import time
import uuid
from typing import Any, Dict, List, Literal, Optional, Tuple, Union
from urllib.parse import unquote, urljoin, urlparse

import pathvalidate
import requests
from Levenshtein import ratio
from pydantic import Field
from termcolor import colored

from tapeagents.config import common_cache_dir, force_cache
from tapeagents.core import Action, Observation
from tapeagents.tools.base import StatefulTool
from tapeagents.tools.converters import (
    FileConversionException,
    FileConverter,
    FileConverterOptions,
    UnsupportedFormatException,
)
from tapeagents.utils import FatalError, diff_strings

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

_CACHE_PREFIX = "web_cache"


class SimpleTextBrowser:
    """A minimal text-based web browser designed for AI agent use."""

    user_agent = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/126.0.0.0 Safari/537.36"

    def __init__(
        self,
        start_page: Optional[str] = None,
        viewport_size: Optional[int] = 32000,
        downloads_folder: str = "/tmp/agent_browser_downloads",
        use_web_cache: bool = True,
        request_kwargs: Optional[Union[Dict[str, Any], None]] = None,
        converter_kwargs: Optional[Dict[str, Any]] = None,
        file_converter_options: Optional[FileConverterOptions] = None,
    ):
        self.start_page: str = start_page if start_page else "about:blank"
        self.viewport_size = viewport_size  # Applies only to the standard uri types
        self.downloads_folder = downloads_folder
        os.makedirs(self.downloads_folder, exist_ok=True)
        self.history: List[Tuple[str, float]] = list()
        self.page_title: Optional[str] = None
        self.viewport_current_page = 0
        self.viewport_pages: List[Tuple[int, int]] = list()
        self.set_address(self.start_page)
        self.request_kwargs = request_kwargs or {"headers": {"User-Agent": self.user_agent}}
        self.request_kwargs["headers"] = self.request_kwargs.get("headers", {})

        self._mdconvert = FileConverter(file_converter_options=file_converter_options)

        self._page_content: str = ""
        self._page_error: int = 0
        self.converter_kwargs = converter_kwargs or {}

        self._find_on_page_query: Union[str, None] = None
        self._find_on_page_last_result: Union[int, None] = None  # Location of the last result

        self.use_web_cache = use_web_cache
        self._cache = {}
        self._log = []
        self._cache_buffer = []
        self.load_cache()

    def load_cache(self):
        cache_dir = common_cache_dir()
        if os.path.exists(cache_dir):
            for fname in os.listdir(cache_dir):
                if not fname.startswith(_CACHE_PREFIX):
                    continue
                with open(os.path.join(cache_dir, fname)) as f:
                    for line in f:
                        data = json.loads(line)
                        self._cache[data["k"]] = data["v"]
        logger.info(f"Loaded {len(self._cache)} web results from cache {cache_dir}")

    @property
    def address(self) -> str:
        """Return the address of the current page."""
        return self.history[-1][0]

    def set_address(self, uri_or_path: str) -> None:
        """Update the address, visit the page, and set the content of the viewport."""
        self.history.append((uri_or_path, time.time()))

        # Handle special URIs
        if uri_or_path == "about:blank":
            self._set_page_content("")
        else:
            if (
                not uri_or_path.startswith("http:")
                and not uri_or_path.startswith("https:")
                and not uri_or_path.startswith("file:")
            ):
                if len(self.history) > 1:
                    prior_address = self.history[-2][0]
                    uri_or_path = urljoin(prior_address, uri_or_path)
                    # Update the address with the fully-qualified path
                    self.history[-1] = (uri_or_path, self.history[-1][1])
            self._fetch_page(uri_or_path)

        self.viewport_current_page = 0
        self.find_on_page_query = None
        self.find_on_page_viewport = None

    @property
    def viewport(self) -> str:
        """Return the content of the current viewport."""
        bounds = self.viewport_pages[self.viewport_current_page]
        return self.page_content[bounds[0] : bounds[1]]

    @property
    def page_content(self) -> str:
        """Return the full contents of the current page."""
        return self._page_content

    def _set_page_content(self, content: str) -> None:
        """Sets the text content of the current page."""
        self._page_content = content
        self._split_pages()
        if self.viewport_current_page >= len(self.viewport_pages):
            self.viewport_current_page = len(self.viewport_pages) - 1

    def page_down(self) -> None:
        self.viewport_current_page = min(self.viewport_current_page + 1, len(self.viewport_pages) - 1)

    def page_up(self) -> None:
        self.viewport_current_page = max(self.viewport_current_page - 1, 0)

    def find_on_page(self, query: str) -> Union[str, None]:
        """Searches for the query from the current viewport forward, looping back to the start if necessary."""

        # Did we get here via a previous find_on_page search with the same query?
        # If so, map to find_next
        if query == self._find_on_page_query and self.viewport_current_page == self._find_on_page_last_result:
            return self.find_next()

        # Ok it's a new search start from the current viewport
        self._find_on_page_query = query
        viewport_match = self._find_next_viewport(query, self.viewport_current_page)
        if viewport_match is None:
            self._find_on_page_last_result = None
            return None
        else:
            self.viewport_current_page = viewport_match
            self._find_on_page_last_result = viewport_match
            return self.viewport

    def find_next(self) -> str | None:
        """Scroll to the next viewport that matches the query"""

        if self._find_on_page_query is None:
            return None

        starting_viewport = self._find_on_page_last_result
        if starting_viewport is None:
            starting_viewport = 0
        else:
            starting_viewport += 1
            if starting_viewport >= len(self.viewport_pages):
                starting_viewport = 0

        viewport_match = self._find_next_viewport(self._find_on_page_query, starting_viewport)
        if viewport_match is None:
            self._find_on_page_last_result = None
            return None
        else:
            self.viewport_current_page = viewport_match
            self._find_on_page_last_result = viewport_match
            return self.viewport

    def _find_next_viewport(self, query: str, starting_viewport: int) -> Union[int, None]:
        """Search for matches between the starting viewport looping when reaching the end."""

        if query is None:
            return None

        # Normalize the query, and convert to a regular expression
        nquery = re.sub(r"\*", "__STAR__", query)
        nquery = " " + (" ".join(re.split(r"\W+", nquery))).strip() + " "
        nquery = nquery.replace(" __STAR__ ", "__STAR__ ")  # Merge isolated stars with prior word
        nquery = nquery.replace("__STAR__", ".*").lower()

        if nquery.strip() == "":
            return None

        idxs = list()
        idxs.extend(range(starting_viewport, len(self.viewport_pages)))
        idxs.extend(range(0, starting_viewport))

        for i in idxs:
            bounds = self.viewport_pages[i]
            content = self.page_content[bounds[0] : bounds[1]]

            # TODO: Remove markdown links and images
            ncontent = " " + (" ".join(re.split(r"\W+", content))).strip().lower() + " "
            if re.search(nquery, ncontent):
                return i

        return None

    def _split_pages(self) -> None:
        # Do not split search results
        if self.address.startswith("search:"):
            self.viewport_pages = [(0, len(self._page_content))]
            return

        # Handle empty pages
        if len(self._page_content) == 0:
            self.viewport_pages = [(0, 0)]
            return

        # Break the viewport into pages
        self.viewport_pages = []
        start_idx = 0
        while start_idx < len(self._page_content):
            end_idx = min(start_idx + self.viewport_size, len(self._page_content))  # type: ignore[operator]
            # Adjust to end on a space
            while end_idx < len(self._page_content) and self._page_content[end_idx - 1] not in [" ", "\t", "\r", "\n"]:
                end_idx += 1
            self.viewport_pages.append((start_idx, end_idx))
            start_idx = end_idx

    def _fetch_page(self, url: str) -> None:
        download_path = ""
        response = None
        try:
            if url.startswith("file://"):
                download_path = os.path.normcase(os.path.normpath(unquote(url[7:])))
                res = self._mdconvert.convert_local(download_path, **self.converter_kwargs)
                self.page_title = res.title
                self._set_page_content(res.text_content)
            else:
                # Prepare the request parameters
                request_kwargs = self.request_kwargs.copy() if self.request_kwargs is not None else {}
                request_kwargs["stream"] = True

                response = requests.get(url, **request_kwargs)
                response.raise_for_status()

                content_type = response.headers.get("content-type", "")
                # Text or HTML
                if "text/" in content_type.lower():
                    res = self._mdconvert.convert_response(response, **self.converter_kwargs)
                    self.page_title = res.title
                    self._set_page_content(res.text_content)
                # A download
                else:
                    # Try producing a safe filename
                    fname = None
                    download_path = None
                    try:
                        fname = pathvalidate.sanitize_filename(os.path.basename(urlparse(url).path)).strip()
                        download_path = os.path.abspath(os.path.join(self.downloads_folder, fname))

                        suffix = 0
                        while os.path.exists(download_path) and suffix < 1000:
                            suffix += 1
                            base, ext = os.path.splitext(fname)
                            new_fname = f"{base}__{suffix}{ext}"
                            download_path = os.path.abspath(os.path.join(self.downloads_folder, new_fname))
                    except NameError:
                        pass

                    # No suitable name, so make one
                    if fname is None:
                        extension = mimetypes.guess_extension(content_type)
                        if extension is None:
                            extension = ".download"
                        fname = str(uuid.uuid4()) + extension
                        download_path = os.path.abspath(os.path.join(self.downloads_folder, fname))

                    # Open a file for writing
                    if not download_path:
                        raise ValueError("Could not determine a suitable download path.")

                    with open(download_path, "wb") as fh:
                        for chunk in response.iter_content(chunk_size=512):
                            fh.write(chunk)

                    # Render it
                    local_uri = pathlib.Path(download_path).as_uri()
                    self.set_address(local_uri)

        except UnsupportedFormatException as e:
            logger.error(colored(f"UnsupportedFormatException: {e}", "red"))
            self.page_title = "Unsupported Format"
            self._set_page_content(f"Unsupported Format File: {e}")
            self._page_error = 1
        except FileConversionException as e:
            logger.error(colored(f"FileConversionException: {e}", "red"))
            self.page_title = "Failed to read file"
            self._set_page_content(f"Error: {e}")
            self._page_error = 2
        except FileNotFoundError:
            self.page_title = "Error 404"
            self._set_page_content(f"## Error 404\n\nFile not found: {download_path}")
            self._page_error = 404
        except requests.exceptions.RequestException as e:
            if response is None:
                self._set_page_content(f"## Error {e}")
                self._page_error = 3
            else:
                self.page_title = f"Error {response.status_code}"
                self._page_error = response.status_code
                # If the error was rendered in HTML we might as well render it
                content_type = response.headers.get("content-type", "")
                if content_type is not None and "text/html" in content_type.lower():
                    res = self._mdconvert.convert(response, **self.converter_kwargs)
                    self.page_title = f"Error {response.status_code}"
                    self._set_page_content(f"## Error {response.status_code}\n\n{res.text_content[:500]}")
                else:
                    text = ""
                    for chunk in response.iter_content(chunk_size=512, decode_unicode=True):
                        text += chunk
                    self.page_title = f"Error {response.status_code}"
                    self._set_page_content(f"## Error {response.status_code}\n\n{text[:500]}")
        except Exception as e:
            self._page_error = -1
            self.page_title = "Failed to fetch page"
            self._set_page_content(str(e))

    def page_with_title(self) -> str:
        if self._page_error:
            header = (
                f"Failed to load page, Error {self._page_error}\nTitle: {self.page_title}\n=======================\n"
            )
        else:
            header = f"Title: {self.page_title}\n=======================\n" if self.page_title else ""
        return header + self.viewport.strip()

    def _add_to_cache(self, key: str, value: Any) -> None:
        self._cache[key] = value
        self._log.append({"k": key, "v": value})
        self._cache_buffer.append({"k": key, "v": value})
        self.flush_cache()

    def flush_cache(self):
        fname = os.path.join(common_cache_dir(), f"{_CACHE_PREFIX}.{os.getpid()}.{threading.get_native_id()}.jsonl")
        with open(fname, "a") as f:
            for item in self._cache_buffer:
                f.write(json.dumps(item) + "\n")
        self._cache_buffer = []

    def flush_log(self, exp_dir: str):
        os.makedirs(os.path.join(exp_dir, "browser_log"), exist_ok=True)
        browser_log_path = os.path.join(exp_dir, f"browser_log/{os.getpid()}.{threading.get_native_id()}.jsonl")
        if len(self._log):
            with open(browser_log_path, "a") as wf:
                for line in self._log:
                    wf.write(json.dumps(line) + "\n")
            self._log = []

    def get_page(self, url: str) -> tuple[str, int, int]:
        """
        Loads a web page and returns a tuple containing:
        - Content of its first viewport (first screen)
        - Current page number
        - Total number of pages
        """
        self._page_error = 0
        if url.startswith("/"):
            # in case of a local file
            url = f"file://{url}"
        if self.use_web_cache and url in self._cache:
            logger.info(colored(f"Cache hit {url}", "green"))
            self._log.append({"k": url, "v": self._cache[url]})
            content, title = self._cache[url]
            self.history.append((url, time.time()))
            self.page_title = title
            self._set_page_content(content)
            self.viewport_current_page = 0
        elif force_cache():
            ratios = [(k, ratio(url, k, score_cutoff=0.7)) for k in self._cache.keys()]
            closest, score = sorted(ratios, key=lambda x: x[1], reverse=True)[0]
            if score >= 0.7:
                logger.warning(diff_strings(url, closest))
                logger.warning(f"Closest url score: {score:.3f}")
            raise FatalError(f"Page {url} not in cache")
        else:
            logger.info(colored(f"Page {url} not in cache", "yellow"))
            self.page_title = ""
            self.set_address(url)
            self._add_to_cache(url, (self.page_content, self.page_title))
        error = self._page_error
        self._page_error = 0
        return (self.page_with_title(), len(self.viewport_pages), error)

    def get_next_page(self) -> tuple[str, int, int]:
        """
        Load next page of the document and return the current content of the viewport, current page number and total number of pages.
        """
        if self.viewport_current_page + 1 == len(self.viewport_pages):
            raise ValueError("No more pages to read.")
        self.page_down()
        return (
            self.page_with_title(),
            self.viewport_current_page + 1,
            len(self.viewport_pages),
        )

    def get_whole_document(self, url: str) -> str:
        try:
            self.get_page(url)
        except Exception as e:
            raise Exception(f"Failed to load page {url}.\nError: {e}")
        return self.page_content


class PageDownAction(Action):
    """
    Action that scrolls down to display the next page of the current document.
    """

    kind: Literal["page_down_action"] = "page_down_action"


class PageUpAction(Action):
    """
    Action that scrolls up to display the previous page of the current document.
    """

    kind: Literal["page_up_action"] = "page_up_action"


class ReadDocumentAction(Action):
    """
    Action that loads a document, file, image, or web page from the provided URL or file path.
    Returns the first page of its content. Use page_down_action to view subsequent pages.
    """

    kind: Literal["read_document_action"] = "read_document_action"
    url: str = Field(description="URL or file path of the document")
    fact_description: str = Field(description="Description of the fact to search for in the document")
    fact_name: str = Field(description="Name of the fact to search for in the document")


class PageObservation(Observation):
    kind: Literal["page_observation"] = "page_observation"
    text: str
    current_page: int
    total_pages: int
    error: int | str | None = None

    def short_view(self, max_chars=100):
        view = self.llm_dict()
        view_json = json.dumps(view, indent=2, ensure_ascii=False)
        if len(view["text"]) > max_chars:
            view["text"] = view["text"][:max_chars] + "..."
            view_json = json.dumps(view, indent=2, ensure_ascii=False)
        return view_json


class SimpleBrowser(StatefulTool):
    """
    Simple text-based web browser for reading documents and web pages.
    Can load web page or document from a URL or file path and scroll through its content.
    """

    actions: tuple[type[Action], ...] = (ReadDocumentAction, PageDownAction)
    observations: tuple[type[Observation], ...] = (PageObservation,)
    exp_path: str
    kwargs: dict[str, Any] | None = None
    _browser: SimpleTextBrowser = None  # type: ignore

    def model_post_init(self, __context: Any):
        kwargs = self.kwargs or {}
        self._browser = SimpleTextBrowser(**kwargs)

    def execute_action(self, action: ReadDocumentAction | PageDownAction) -> PageObservation:
        if isinstance(action, ReadDocumentAction):
            text, total_pages, error = self._browser.get_page(action.url)
            obs = PageObservation(text=text, current_page=1, total_pages=total_pages, error=error or None)
        else:
            text, current_page, total_pages = self._browser.get_next_page()
            obs = PageObservation(
                text=text,
                current_page=current_page,
                total_pages=total_pages,
                error=self._browser._page_error if self._browser._page_error else None,
            )
        return obs

    def close(self) -> None:
        self._browser.flush_log(self.exp_path)



import asyncio
import json
import logging
import os
import time
from contextlib import AsyncExitStack
from datetime import timedelta
from typing import Any, Optional

import nest_asyncio
from mcp import ClientSession, StdioServerParameters, Tool as MCPTool, stdio_client
from mcp.types import CallToolResult, TextContent

from tapeagents.config import force_cache
from tapeagents.core import Action, LLMOutputParsingFailureAction, Observation
from tapeagents.environment import ToolCollectionEnvironment
from tapeagents.tool_calling import FunctionSpec, ToolCallAction, ToolResult, ToolSpec
from tapeagents.tools.base import BaseTool
from tapeagents.tools.tool_cache import add_to_cache, get_from_cache
from tapeagents.utils import FatalError

logger = logging.getLogger(__name__)


class NoTool(Exception):
    """Raised when a tool is not found in the MCP client"""

    pass


class MCPClient:
    def __init__(self, config_path: str, use_cache: bool = False, read_timeout_seconds: int = 10) -> None:
        self.servers = self.load_config(config_path)
        self.sessions: dict[str, ClientSession] = {}
        self.tools: dict[str, MCPTool] = {}
        self.tool_to_server: dict[str, str] = {}
        self.use_cache = use_cache
        self.read_timeout_seconds = read_timeout_seconds
        self.exit_stack = AsyncExitStack()

    async def start_servers(self):
        for server_name, server_params in self.servers.items():
            stdio_transport = await self.exit_stack.enter_async_context(stdio_client(server_params))
            stdio, write = stdio_transport
            session = await self.exit_stack.enter_async_context(
                ClientSession(stdio, write, read_timeout_seconds=timedelta(seconds=self.read_timeout_seconds))
            )
            await session.initialize()
            self.sessions[server_name] = session
            response = await session.list_tools()
            for tool in response.tools:
                if tool.name in self.tools:
                    raise Exception(
                        f"Tools conflict! Tool {tool.name} already provided by server '{self.tool_to_server[tool.name]}'"
                    )
                self.tools[tool.name] = tool
                self.tool_to_server[tool.name] = server_name
            logger.info(f"Connected to MCP server '{server_name}' with tools: {[tool.name for tool in response.tools]}")
        logger.info(f"Started {len(self.servers)} MCP servers")

    def load_config(self, config_path) -> dict[str, StdioServerParameters]:
        assert os.path.exists(config_path), f"Config path {config_path} does not exist"
        self.config_path = config_path

        try:
            with open(config_path, "r") as f:
                self.config = json.load(f)
        except json.JSONDecodeError as e:
            raise ValueError(f"Failed to parse {config_path}, invalid json: {e}")
        try:
            server_configs: dict[str, dict] = self.config["mcpServers"]
            assert isinstance(server_configs, dict), "mcpServers must be a dict"
            assert len(server_configs) > 0, "mcpServers dict is empty"
        except Exception as e:
            raise ValueError(f"Failed to get MCP server configs from {config_path}: {e}")

        servers: dict[str, StdioServerParameters] = {}
        for server_name, server_config_dict in server_configs.items():
            try:
                server_config_dict = self.prepare_env_vars(server_config_dict)
                server_params = StdioServerParameters.model_validate(server_config_dict)
            except Exception as e:
                raise ValueError(f"Failed to parse server config {server_config_dict}: {e}")
            servers[server_name] = server_params
        logger.info(f"Loaded {len(servers)} MCP server configs from {config_path}")
        return servers

    def prepare_env_vars(self, server_config_dict: dict) -> dict:
        if server_env := server_config_dict.get("env"):
            for env_var, env_value in server_env.items():
                if env_var in os.environ and not env_value:  # reuse existing env var value if not set in config
                    logger.info(f"Set mcp server env var {env_var} from current environment")
                    server_config_dict["env"][env_var] = os.environ[env_var]
        return server_config_dict

    async def call_tool(self, tool_name: str, tool_args: dict[str, Any]) -> CallToolResult:
        server_name = self.check_tool_exists(tool_name)
        # Current implementation of cache assumes tool calls are deterministic and do not alter state
        if self.use_cache:
            tool_name_key = f"mcp.{server_name}.{tool_name}"
            result = get_from_cache(tool_name_key, args=(), kwargs=tool_args)
            if not result and force_cache():
                raise FatalError(f"Cache is forced but no cache entry found for {tool_name_key}({tool_args})")
            if not result:
                result = await self._call_tool(server_name, tool_name, tool_args)
                add_to_cache(tool_name_key, args=(), kwargs=tool_args, result=result.model_dump(exclude_none=True))
            else:
                result = CallToolResult(**result)
        else:
            result = await self._call_tool(server_name, tool_name, tool_args)

        return result

    async def _call_tool(self, server_name: str, tool_name: str, tool_args: dict[str, Any]) -> CallToolResult:
        try:
            session = self.sessions[server_name]
            result = await session.call_tool(tool_name, tool_args)
        except Exception as e:
            logger.exception(f"Error calling tool {tool_name}: {e}")
            raise e
        return result

    def check_tool_exists(self, tool_name):
        try:
            server_name = self.tool_to_server[tool_name]
        except KeyError:
            raise NoTool(f"Tool {tool_name} not found in any of the MCP servers")
        return server_name

    async def close(self) -> None:
        await self.exit_stack.aclose()


class MCPEnvironment(ToolCollectionEnvironment):
    def __init__(
        self,
        config_path: str = "",
        tools_whitelist: Optional[list[str]] = None,
        tools: Optional[list[BaseTool]] = None,
        use_cache: bool = False,
        read_timeout_seconds: int = 10,
        client: MCPClient | None = None,
    ) -> None:
        super().__init__(tools=tools or [])
        logger.info(f"Initializing MCPEnvironment with config_path: {config_path}")
        self.client = client or MCPClient(
            config_path=config_path, use_cache=use_cache, read_timeout_seconds=read_timeout_seconds
        )
        self.tools_whitelist = tools_whitelist or []

        if not self.client and not self.tools:
            raise ValueError("Tools or MCP client config_path must be provided")
        self.client.use_cache = use_cache

    def initialize(self):
        nest_asyncio.apply()
        self.loop = asyncio.get_event_loop()
        self.loop.run_until_complete(self.client.start_servers())
        self.tools.extend(
            [
                ToolSpec(
                    function=FunctionSpec(
                        name=tool.name, description=tool.description or "", parameters=tool.inputSchema
                    )
                )
                for tool in self.client.tools.values()
                if tool.name in self.tools_whitelist or not self.tools_whitelist
            ]
        )

    async def ainitialize(self) -> None:
        await self.client.start_servers()
        self.tools.extend(
            [
                ToolSpec(
                    function=FunctionSpec(
                        name=tool.name, description=tool.description or "", parameters=tool.inputSchema
                    )
                )
                for tool in self.client.tools.values()
                if tool.name in self.tools_whitelist or not self.tools_whitelist
            ]
        )

    def actions(self) -> tuple[type[Action] | ToolSpec, ...]:
        actions = super().actions()
        tool_specs = [t for t in self.tools if isinstance(t, ToolSpec)]
        return actions + tuple(tool_specs)

    def step(self, action: Action) -> Observation:
        if not isinstance(action, ToolCallAction):
            return super().step(action)
        if isinstance(action, LLMOutputParsingFailureAction):
            return ToolResult(tool_call_id="", content="Try again")
        try:
            assert self.client is not None, "MCPClient is not initialized"
            result = self.loop.run_until_complete(
                self.client.call_tool(action.function.name, action.function.arguments)
            )
        except NoTool:
            logger.exception(f"Tool {action.function.name} not found in MCP client")
            result = CallToolResult(
                content=[TextContent(type="text", text=f"Tool {action.function.name} not found")], isError=True
            )
        except KeyError as e:
            logger.exception(f"KeyError when executing MCP tool call: {e}")
            result = CallToolResult(
                content=[TextContent(type="text", text=f"Error executing tool {action.function.name}: KeyError {e}")],
                isError=True,
            )
        except Exception as e:
            logger.exception(f"Error executing MCP tool call: {e}")
            result = CallToolResult(
                content=[TextContent(type="text", text=f"Error executing tool {action.function.name}: {str(e)}")],
                isError=True,
            )
        return ToolResult(tool_call_id=action.id, content=result)

    def close(self) -> None:
        super().close()
        if self.client is not None:
            try:
                self.loop.run_until_complete(self.client.close())
            except Exception as e:
                logger.warning(f"Failed to close MCP client properly: {e}")

    async def astep(self, action: Action) -> Observation:
        if not isinstance(action, ToolCallAction):
            return await super().astep(action)
        if isinstance(action, LLMOutputParsingFailureAction):
            return ToolResult(tool_call_id="", content="Try again")
        t = time.perf_counter()
        try:
            assert self.client is not None, "MCPClient is not initialized"
            result = await self.client.call_tool(action.function.name, action.function.arguments)
        except NoTool:
            logger.exception(f"Tool {action.function.name} not found in MCP client")
            result = CallToolResult(
                content=[TextContent(type="text", text=f"Tool {action.function.name} not found")], isError=True
            )
        except KeyError as e:
            logger.exception(f"KeyError when executing MCP tool call: {e}")
            result = CallToolResult(
                content=[TextContent(type="text", text=f"Error executing tool {action.function.name}: KeyError {e}")],
                isError=True,
            )
        except Exception as e:
            logger.exception(f"Error executing MCP tool call: {e}")
            result = CallToolResult(
                content=[TextContent(type="text", text=f"Error executing tool {action.function.name}: {str(e)}")],
                isError=True,
            )
        observation = ToolResult(tool_call_id=action.id, content=result)
        observation.metadata.other["action_execution_time"] = time.perf_counter() - t
        observation.metadata.other["action_kind"] = action.kind
        return observation

    async def aclose(self) -> None:
        await super().aclose()
        await self.client.close()



import json
import logging
import os
import sys

from tapeagents.io import load_tapes

from ..eval import get_exp_config_dict, load_dataset, tape_correct
from ..steps import GaiaTape

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)


def main(exp_path: str):
    assert os.path.isdir(exp_path), f"Directory {exp_path} does not exist or is not a directory"
    cfg = get_exp_config_dict(exp_path)
    tasks = load_dataset(cfg["split"])
    submission_file = os.path.join(exp_path, "submission.jsonl")
    submission = []
    expected_tasks_num = sum(len(level_tasks) for level_tasks in tasks.values())
    total_unsolved = 0
    all_tapes = load_tapes(GaiaTape, os.path.join(exp_path, "tapes"), file_extension=".json")  # type: ignore
    tapes_by_task_id = {tape.metadata.task["task_id"]: tape for tape in all_tapes}
    print(f"Loaded {len(all_tapes)} tapes")
    correct = 0
    for level, level_tasks in tasks.items():
        unsolved = 0
        tapes = []
        for task in level_tasks:
            tape = tapes_by_task_id.get(task["task_id"])
            last_step = tape.steps[-1]
            model_answer = last_step.answer if last_step.kind == "gaia_answer_action" else None
            if model_answer != tape.metadata.result:
                print(f"Model answer '{model_answer}' does not match tape result '{tape.metadata.result}'")
            if not model_answer:
                unsolved += 1
                model_answer = "0"
            if tape_correct(tape):
                correct += 1
            line = {"task_id": tape.metadata.task["task_id"], "model_answer": str(model_answer)}
            submission.append(line)
            tapes.append(tape)
        logger.info(f"Level {level+1}: gathered {len(tapes)} task solutions, {unsolved} unsolved, {correct} correct")
        total_unsolved += unsolved

    if len(submission) != expected_tasks_num:
        logger.info(f"Submission is not ready, expected {expected_tasks_num} task tapes, got {len(submission)}")
        return None
    else:
        logger.info(f"Submission is ready, has {total_unsolved} unsolved tasks out of {expected_tasks_num}")
        logger.info(f"Writing submission with {len(submission)} tasks to {submission_file}")
        with open(submission_file, "w") as f:
            for line in submission:
                f.write(f"{json.dumps(line)}\n")
        return submission_file


if __name__ == "__main__":
    assert len(sys.argv) == 2, "Usage: examples.gaia_agent.scripts.prepare_submission <exp_dir>"
    submission_file = main(sys.argv[1])



import asyncio
import logging
import os
import time

import aiohttp
import hydra
from omegaconf import DictConfig

from examples.gaia_agent.eval import load_dataset, tape_correct, task_to_observations
from examples.gaia_agent.steps import GaiaTape
from tapeagents.core import StopStep
from tapeagents.io import save_json_tape
from tapeagents.orchestrator import run_agent_with_remote_env

logger = logging.getLogger(__name__)


@hydra.main(
    version_base=None,
    config_path=".",
    config_name="async_remote",
)
def main(cfg: DictConfig) -> None:
    asyncio.run(amain(cfg))


async def amain(cfg: DictConfig) -> None:
    cfg.llm.base_url = os.environ["BASE_URL"]
    tasks = load_dataset(cfg.split)

    dt = time.perf_counter()
    timeout = 3600.0
    connector = aiohttp.TCPConnector(limit=1000, limit_per_host=1000, force_close=True)
    timeout = aiohttp.ClientTimeout(total=timeout, connect=timeout, sock_read=timeout)
    coroutines = []
    results = []
    if cfg.only_tasks:
        selected_tasks: list[tuple[int, int]] = [(level, task_num) for level, task_num in cfg.only_tasks]
    else:
        selected_tasks = [
            (level, task_num) for level, level_tasks in tasks.items() for task_num, task in enumerate(level_tasks)
        ]

    async with aiohttp.ClientSession(connector=connector, timeout=timeout) as session:
        for level, task_num in selected_tasks:
            logging.info(f"Schedule task {level=}, {task_num=}")
            task = tasks[level][task_num]
            start_tape = GaiaTape(steps=task_to_observations(task))  # type: ignore
            start_tape.metadata.id = f"l{level}_task{task_num:03d}"
            start_tape.metadata.task = task
            coroutines.append(run_agent_with_remote_env(cfg, start_tape, session))
        logger.info(f"Solving {len(coroutines)} tasks")
        results = await asyncio.gather(*coroutines)

    total_steps = 0
    tapes_with_errors = 0
    solved = 0
    for tape in results:
        stop_steps = [step for step in tape if isinstance(step, StopStep)]
        tape.metadata.result = "" if not stop_steps else stop_steps[-1].model_dump().get("answer", "")
        solved += int(tape_correct(tape))
        total_steps += len(tape.steps)
        if tape.metadata.error:
            tapes_with_errors += 1
    logger.info(f"Average tape length: {(total_steps / len(results)):.2f}")
    logger.info(f"Tapes with error: {(tapes_with_errors / len(results)):.2f} ({tapes_with_errors} of {len(results)})")
    logger.info(f"Accuracy: {(solved / len(results)):.2f} ({solved} of {len(results)})")
    for tape in results:
        save_json_tape(tape, os.path.join(cfg.exp_path, "tapes"), tape.metadata.parent_id)
    logger.info(f"Execution time: {time.perf_counter() - dt:.2f} seconds")


if __name__ == "__main__":
    print("Make sure that you run environment server first using `uv run examples/workarena/env_server.py -cp ../mcp/`")
    main()



import base64
import os
from io import BytesIO

from openai import OpenAI
from PIL import Image


class GroundingModel:
    def __init__(self, url: str = "", api_key: str = ""):
        if not url:
            try:
                url = os.environ["GROUNDING_API_URL"]
            except KeyError:
                raise RuntimeError("GROUNDING_API_URL environment variable is not set")
        if not api_key:
            try:
                api_key = os.popen("eai login --profile default token").read().strip()
                if not api_key:
                    api_key = os.getenv("GROUNDING_API_KEY")
            except Exception:
                api_key = os.getenv("GROUNDING_API_KEY")
            if not api_key:
                raise RuntimeError(
                    "Failed to get API key from 'eai login token' or GROUNDING_API_KEY environment variable"
                )
        self.vl_client = OpenAI(
            base_url=f"{url}/v1",
            api_key=api_key,
        )
        self.model = "osunlp/UGround-V1-7B"

    def get_coords(self, image: Image, element_description: str):
        img_width, img_height = image.size
        buffered = BytesIO()
        image.save(buffered, format="PNG")
        base64_image = base64.b64encode(buffered.getvalue()).decode()
        messages = self.format_openai_template(f"click {element_description}", base64_image)
        completion = self.vl_client.chat.completions.create(model=self.model, messages=messages, temperature=0.0)
        out_text = completion.choices[0].message.content
        str_x, str_y = out_text.strip("()").split(",")
        x = int(str_x.strip())
        y = int(str_y.strip())
        real_x = x / 1000 * img_width
        real_y = y / 1000 * img_height
        return real_x, real_y

    def format_openai_template(self, description: str, base64_image):
        return [
            {
                "role": "user",
                "content": [
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"},
                    },
                    {
                        "type": "text",
                        "text": f"""
    Your task is to help the user identify the precise coordinates (x, y) of a specific area/element/object on the screen based on a description.

    - Your response should aim to point to the center or a representative point within the described area/element/object as accurately as possible.
    - If the description is unclear or ambiguous, infer the most relevant area or element based on its likely context or purpose.
    - Your answer should be a single string (x, y) corresponding to the point of the interest.

    Description: {description}

    Answer:""",
                    },
                ],
            },
        ]
